from uqbar.containers import UniqueTreeContainer
from uqbar.graphs.Attributes import Attributes
from uqbar.graphs.Edge import Edge  # noqa
from uqbar.graphs.Node import Node  # noqa
from typing import Dict, List, Mapping, Set, Tuple, Union  # noqa


class Graph(UniqueTreeContainer):
    """
    A Graphviz graph, subgraph or cluster.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Core Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Union[Mapping[str, object], Attributes]=None,
        edge_attributes: Union[Mapping[str, object], Attributes]=None,
        is_cluster: bool=False,
        is_digraph: bool=True,
        name: str=None,
        node_attributes: Union[Mapping[str, object], Attributes]=None
        ) -> None:
        UniqueTreeContainer.__init__(self, name=name, children=children)
        self._attributes = Attributes(
            'cluster' if is_cluster else 'graph',
            **(attributes or {}),
            )
        self._edge_attributes = Attributes('edge', **(edge_attributes or {}))
        self._node_attributes = Attributes('node', **(node_attributes or {}))
        self._is_cluster = bool(is_cluster)
        self._is_digraph = bool(is_digraph)

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        def recurse(graph):
            indent = '    '
            result = []
            if not graph.parent:
                name = graph.name or 'G'
                if graph.is_digraph:
                    string = 'digraph {} {{'.format(
                        Attributes._format_value(name))
                else:
                    string = 'graph {} {{'.format(
                        Attributes._format_value(name))
            else:
                if graph.name is not None:
                    name = graph.name
                    if graph.is_cluster:
                        name = 'cluster_{}'.format(name)
                else:
                    name = graph._get_canonical_name()
                string = 'subgraph {} {{'.format(
                    Attributes._format_value(name))
            result.append(string)
            if graph.attributes:
                attributes = 'graph {}'.format(
                    format(graph.attributes, 'graphviz')).split('\n')
                result.extend(indent + line for line in attributes)
            if graph.node_attributes:
                attributes = 'node {}'.format(
                    format(graph.node_attributes, 'graphviz')).split('\n')
                result.extend(indent + line for line in attributes)
            if graph.edge_attributes:
                attributes = 'edge {}'.format(
                    format(graph.edge_attributes, 'graphviz')).split('\n')
                result.extend(indent + line for line in attributes)
            for child in graph:
                if isinstance(child, type(self)):
                    lines = (indent + line for line in recurse(child))
                else:
                    lines = (
                        indent + line for line in
                        format(child, 'graphviz').split('\n')
                        )
                result.extend(lines)
            if graph in edge_parents:
                for edge in edge_parents[graph]:
                    lines = format(edge, 'graphviz').split('\n')
                    result.extend(indent + line for line in lines)
            result.append('}')
            return result

        all_edges: Set[Edge] = set()
        for child in self.depth_first():
            for edge in getattr(child, 'edges', ()):
                if edge.tail.root is not edge.head.root:
                    continue
                all_edges.add(edge)
        edge_parents: Dict[Graph, List[Edge]] = {}
        for edge in sorted(
            all_edges,
            key=lambda edge: (edge.tail_graph_order, edge.head_graph_order),
            ):
            highest_parent = edge._get_highest_parent()
            edge_parents.setdefault(highest_parent, []).append(edge)

        return '\n'.join(recurse(self))

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        name_prefix = 'graph'
        if self.is_cluster:
            name_prefix = 'cluster'
        if self.name is not None:
            name = self.name
            root = self.root
            if root:
                instances = root[self.name]
                if not isinstance(instances, type(self)):
                    name = '{}_{}'.format(name, instances.index(self))
            suffix = name
        elif self.graph_order:
            suffix = '_'.join(str(x) for x in self.graph_order)
        else:
            suffix = '0'
        return '{}_{}'.format(name_prefix, suffix)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.Graph,
            uqbar.graphs.Node,
            )

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def edge_attributes(self) -> Attributes:
        return self._edge_attributes

    @property
    def is_cluster(self) -> bool:
        return self._is_cluster

    @property
    def is_digraph(self) -> bool:
        return self._is_digraph

    @property
    def node_attributes(self) -> Attributes:
        return self._node_attributes



import textwrap
from uqbar.strings import normalize
from uqbar.apis.ModuleDocumenter import ModuleDocumenter
from uqbar.apis.RootDocumenter import RootDocumenter
from sphinx.ext.autosummary import extract_summary  # type: ignore
from sphinx.util.docutils import new_document  # type: ignore


class SummarizingRootDocumenter(RootDocumenter):
    """
    A summarizing root documenter.

    Like the :py:class:`~uqbar.apis.SummarizingClassDocumenter` and
    :py:class:`~uqbar.apis.SummarizingModuleDocumenter`, this subclass of
    :py:class:`~uqbar.apis.RootDocumenter` generates root API documentation
    with additional information and organization.

    Each visited module or package receives its own section with members
    organized by their documentation sections and listed in autosummary tables.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.SummarizingRootDocumenter(
        ...     module_documenters=[
        ...         uqbar.apis.ModuleDocumenter('uqbar.io'),
        ...         uqbar.apis.ModuleDocumenter('uqbar.strings'),
        ...         ],
        ...     )
        >>> print(str(documenter))
        API
        ===
        <BLANKLINE>
        .. toctree::
           :hidden:
        <BLANKLINE>
           uqbar/io/index
           uqbar/strings
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: :ref:`uqbar.io <uqbar--io>`
           :class: section-header
        <BLANKLINE>
        Tools for IO and file-system manipulation.
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: subsection-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~uqbar.io.find_common_prefix
           ~uqbar.io.find_executable
           ~uqbar.io.relative_to
           ~uqbar.io.walk
           ~uqbar.io.write
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: :ref:`uqbar.strings <uqbar--strings>`
           :class: section-header
        <BLANKLINE>
        Tools for string manipulation.
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: subsection-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~uqbar.strings.delimit_words
           ~uqbar.strings.normalize
           ~uqbar.strings.to_dash_case
           ~uqbar.strings.to_snake_case

    :param module_documenters: a list of of documenters for modules and
        packages of the root documenter; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    def __str__(self):
        result = [
            self.title,
            '=' * len(self.title),
            '',
            '.. toctree::',
            '   :hidden:',
            '',
            ]
        for documenter in self.module_documenters:
            path = documenter.package_path.replace('.', '/')
            if documenter.is_package:
                path += '/index'
            result.append('   {}'.format(path))
        for module_documenter, documenters_by_section in self._recurse(self):
            result.extend([
                '',
                '.. raw:: html',
                '',
                '   <hr/>',
                '',
                '.. rubric:: :ref:`{} <{}>`'.format(
                    module_documenter.package_path,
                    module_documenter.reference_name,
                    ),
                '   :class: section-header',
                ])
            summary = self._extract_summary(module_documenter)
            if summary:
                result.extend(['', summary])
            for section_name, documenters in documenters_by_section:
                result.extend([
                    '',
                    '.. raw:: html',
                    '',
                    '   <hr/>',
                    '',
                    '.. rubric:: {}'.format(section_name),
                    '   :class: subsection-header',
                    '',
                    '.. autosummary::',
                    '   :nosignatures:',
                    '',
                    ])
                for documenter in documenters:
                    result.append('   ~{}'.format(documenter.package_path))
        return '\n'.join(result)

    def _recurse(self, documenter):
        result = []
        if (
            isinstance(documenter, ModuleDocumenter) and
            not documenter.is_nominative
            ):
            result.append((
                documenter,
                documenter.member_documenters_by_section,
                ))
        for module_documenter in documenter.module_documenters:
            result.extend(self._recurse(module_documenter))
        return result

    @classmethod
    def _extract_summary(cls, documenter):
        lines = normalize(documenter.client.__doc__ or '').splitlines()
        if not lines:
            return
        document = new_document('')
        summary = extract_summary(lines, document)
        return '\n'.join(textwrap.wrap(summary, 79))



from typing import Mapping, Tuple, Union
from uqbar.graphs.Attributes import Attributes
from uqbar.containers import UniqueTreeContainer


class Table(UniqueTreeContainer):
    """
    A Graphviz HTML table.

    ::

        >>> import uqbar.graphs
        >>> table = uqbar.graphs.Table([
        ...     uqbar.graphs.TableRow([
        ...         uqbar.graphs.TableCell('Cell One'),
        ...     ]),
        ...     uqbar.graphs.HRule(),
        ...     uqbar.graphs.TableRow([
        ...         uqbar.graphs.TableCell('Cell Two'),
        ...     ]),
        ... ])
        >>> print(format(table, 'graphviz'))
        <TABLE>
            <TR>
                <TD>Cell One</TD>
            </TR>
            <HR/>
            <TR>
                <TD>Cell Two</TD>
            </TR>
        </TABLE>

    ::

        >>> table = uqbar.graphs.Table(
        ...     attributes={
        ...         'border': 5,
        ...         'bgcolor': 'blue',
        ...         },
        ...     )
        >>> print(format(table, 'graphviz'))
        <TABLE BGCOLOR="blue" BORDER="5"></TABLE>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'HTML Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Union[Mapping[str, object], Attributes]=None,
        name: str=None
        ) -> None:
        UniqueTreeContainer.__init__(
            self,
            children=children,
            name=name,
            )
        self._attributes = Attributes('table', **(attributes or {}))

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        start, stop = '<TABLE', '</TABLE>'
        attributes = format(self._attributes, 'html')
        if attributes:
            start += ' {}'.format(attributes)
        start += '>'
        result.append(start)
        for child in self:
            for line in format(child, 'graphviz').splitlines():
                result.append('    {}'.format(line))
        result.append(stop)
        join_character = ''
        if len(result) > 2:
            join_character = '\n'
        return join_character.join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.TableRow,
            uqbar.graphs.HRule,
            )



import uqbar.containers


class PackageNode(uqbar.containers.UniqueTreeContainer):
    """
    A :py:class:`~uqbar.containers.UniqueTreeContainer` subclass used during
    API construction to proxy packages.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Internals'

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        name=None,
        documenter=None,
        source_path=None,
    ):
        super().__init__(children=children, name=name)
        self._source_path = source_path
        self._documenter = documenter

    ### SPECIAL METHODS ###

    def __str__(self):
        result = ['{}/'.format(self.name)]
        for child in self:
            result.extend('    ' + line for line in str(child).splitlines())
        return '\n'.join(result)

    ### PUBLIC PROPERTIES ###

    @property
    def _node_class(self):
        import uqbar.apis
        return (
            uqbar.apis.ModuleNode,
            uqbar.apis.PackageNode,
            )

    @property
    def documenter(self):
        return self._documenter

    @documenter.setter
    def documenter(self, documenter):
        self._documenter = documenter

    @property
    def package_path(self):
        return self._name

    @property
    def source_path(self):
        return self._source_path

    @source_path.setter
    def source_path(self, source_path):
        self._source_path = source_path



import pathlib
import shutil
import uqbar.io
from typing import Sequence, Type, Union
from uqbar.apis.ClassDocumenter import ClassDocumenter
from uqbar.apis.FunctionDocumenter import FunctionDocumenter
from uqbar.apis.MemberDocumenter import MemberDocumenter
from uqbar.apis.ModuleDocumenter import ModuleDocumenter
from uqbar.apis.RootDocumenter import RootDocumenter
from uqbar.apis.ModuleNode import ModuleNode
from uqbar.apis.PackageNode import PackageNode


class APIBuilder(object):
    """
    A builder of reStructuredText API documentation for Python packages.

    ::

        >>> import uqbar
        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as target_directory:
        ...     with uqbar.io.DirectoryChange(target_directory):
        ...         builder = uqbar.apis.APIBuilder(
        ...             initial_source_paths=uqbar.apis.__path__,
        ...             target_directory=target_directory,
        ...             )
        ...         visited_paths = builder()
        ...
        wrote: index.rst
        wrote: uqbar/index.rst
        wrote: uqbar/apis/index.rst
        wrote: uqbar/apis/APIBuilder.rst
        wrote: uqbar/apis/ClassDocumenter.rst
        wrote: uqbar/apis/FunctionDocumenter.rst
        wrote: uqbar/apis/InheritanceGraph.rst
        wrote: uqbar/apis/MemberDocumenter.rst
        wrote: uqbar/apis/ModuleDocumenter.rst
        wrote: uqbar/apis/ModuleNode.rst
        wrote: uqbar/apis/PackageNode.rst
        wrote: uqbar/apis/RootDocumenter.rst
        wrote: uqbar/apis/SummarizingClassDocumenter.rst
        wrote: uqbar/apis/SummarizingModuleDocumenter.rst
        wrote: uqbar/apis/SummarizingRootDocumenter.rst

    :param initial_source_paths: a list of paths to scan for Python sources
    :param target_directory: where to write reStructuredText output
    :param document_private_members: whether to documenter private module members
    :param document_private_modules: whether to document private modules
    :param member_documenter_classes: a list of
        :py:class:`~uqbar.apis.MemberDocumenter` subclasses, defining what classes
        to use to identify and document module members
    :param module_documenter_class: a :py:class:`~uqbar.apis.ModuleDocumenter` subclass
    :param root_documenter_class: a :py:class:`~uqbar.apis.RootDocumenter` subclass
    """

    ### INITIALIZER ###

    def __init__(
        self,
        initial_source_paths: Sequence[Union[str, pathlib.Path]],
        target_directory: Union[str, pathlib.Path],
        document_empty_modules: bool=True,
        document_private_members: bool=False,
        document_private_modules: bool=False,
        member_documenter_classes: Sequence[Type[MemberDocumenter]]=None,
        module_documenter_class: Type[ModuleDocumenter]=None,
        root_documenter_class: Type[RootDocumenter]=None,
        title: str='API',
        ) -> None:
        assert initial_source_paths
        assert target_directory
        self._initial_source_paths = frozenset(
            pathlib.Path(_).resolve().absolute() for _ in initial_source_paths
            )
        self._target_directory = pathlib.Path(target_directory).absolute()
        self._document_private_members = bool(document_private_members)
        self._document_private_modules = bool(document_private_modules)
        self._document_empty_modules = bool(document_empty_modules)
        if member_documenter_classes is None:
            member_documenter_classes = [ClassDocumenter, FunctionDocumenter]
        for _ in member_documenter_classes:
            assert issubclass(_, MemberDocumenter), _
        self._member_documenter_classes = tuple(member_documenter_classes)
        if module_documenter_class is None:
            module_documenter_class = ModuleDocumenter
        assert issubclass(module_documenter_class, ModuleDocumenter)
        self._module_documenter_class = module_documenter_class
        if root_documenter_class is None:
            root_documenter_class = RootDocumenter
        assert issubclass(root_documenter_class, RootDocumenter)
        self._root_documenter_class = root_documenter_class
        self._title = title

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Generate documentation.
        """
        import uqbar.apis
        # Make sure target directory exists.
        if not self._target_directory.exists():
            self._target_directory.mkdir(parents=True)
        self._target_directory = self._target_directory.resolve()
        # What files to document?
        source_paths = uqbar.apis.collect_source_paths(self._initial_source_paths)
        # Build the node tree
        node_tree = self.build_node_tree(source_paths)
        # Get the documenters, in depth-first order.
        documenters = self.collect_module_documenters(node_tree)
        # Execute each documenter, writing to the target directory.
        visited_paths = self.write(documenters)
        # Iterate the target directory, removing un-touched files.
        self.prune(visited_paths)
        return visited_paths

    ### PRIVATE METHODS ###

    def _print(self, message, path):
        cwd = pathlib.Path.cwd()
        if str(path).startswith(str(cwd)):
            path = path.relative_to(cwd)
        print('{} {}'.format((message + ':').ljust(11), path))

    ### PUBLIC METHODS ###

    def build_node_tree(self, source_paths):
        """
        Build a node tree.
        """
        import uqbar.apis
        root = PackageNode()
        # Build node tree, top-down
        for source_path in sorted(
            source_paths,
            key=lambda x: uqbar.apis.source_path_to_package_path(x),
        ):
            package_path = uqbar.apis.source_path_to_package_path(source_path)
            parts = package_path.split('.')
            if (
                not self.document_private_modules and
                any(part.startswith('_') for part in parts)
            ):
                continue
            # Find parent node.
            parent_node = root
            if len(parts) > 1:
                parent_package_path = '.'.join(parts[:-1])
                try:
                    parent_node = root[parent_package_path]
                except KeyError:
                    parent_node = root
                try:
                    if parent_node is root:
                        # Backfill missing parent node.
                        grandparent_node = root
                        if len(parts) > 2:
                            grandparent_node = root[
                                parent_package_path.rpartition('.')[0]
                                ]
                        parent_node = PackageNode(name=parent_package_path)
                        grandparent_node.append(parent_node)
                        grandparent_node[:] = sorted(
                            grandparent_node,
                            key=lambda x: x.package_path,
                            )
                except KeyError:
                    parent_node = root
            # Create or update child node.
            node_class = ModuleNode
            if source_path.name == '__init__.py':
                node_class = PackageNode
            try:
                # If the child exists, it was previously backfilled.
                child_node = root[package_path]
                child_node.source_path = source_path
            except KeyError:
                # Otherwise it needs to be created and appended to the parent.
                child_node = node_class(
                    name=package_path,
                    source_path=source_path,
                    )
                parent_node.append(child_node)
                parent_node[:] = sorted(
                    parent_node,
                    key=lambda x: x.package_path,
                    )
        # Build documenters, bottom-up.
        # This allows parent documenters to easily aggregate their children.
        for node in root.depth_first(top_down=False):
            kwargs = dict(
                document_private_members=self.document_private_members,
                member_documenter_classes=self.member_documenter_classes,
                )
            if isinstance(node, ModuleNode):
                node.documenter = self.module_documenter_class(
                    node.package_path,
                    **kwargs
                    )
            else:
                # Collect references to child modules and packages.
                node.documenter = self.module_documenter_class(
                    node.package_path,
                    module_documenters=[
                        child.documenter for child in node
                        if child.documenter is not None
                        ],
                    **kwargs
                    )
            if (
                not self.document_empty_modules and
                not node.documenter.module_documenters and
                not node.documenter.member_documenters
                ):
                node.parent.remove(node)
        return root

    def collect_module_documenters(self, root_node):
        # Yield root documenter.
        yield self.root_documenter_class(
            module_documenters=[node.documenter for node in root_node],
            title=self._title,
            )
        # Yield module documenters, top-down.
        for node in root_node.depth_first():
            if node is not root_node:
                yield node.documenter

    def write(self, documenters):
        visited_paths = set()
        for documenter in documenters:
            new_documentation = str(documenter)
            path = self._target_directory.joinpath(
                documenter.documentation_path
                )
            cwd = pathlib.Path.cwd()
            if str(path).startswith(str(cwd)):
                path = path.relative_to(pathlib.Path.cwd())
            uqbar.io.write(new_documentation, path, verbose=True)
            visited_paths.add(path.absolute())
        return visited_paths

    def prune(self, visited_paths):
        generator = uqbar.io.walk(self._target_directory, top_down=False)
        for root_path, directory_paths, file_paths in generator:
            for file_path in file_paths[:]:
                if file_path.suffix != '.rst':
                    continue
                if file_path not in visited_paths:
                    file_paths.remove(file_path)
                    file_path.unlink()
                    self._print('pruned', file_path)
            if not file_paths and not directory_paths:
                if root_path == self._target_directory:
                    continue
                shutil.rmtree(str(root_path))
                self._print('pruned', root_path)

    ### PUBLIC PROPERTIES ###

    @property
    def document_empty_modules(self):
        return self._document_empty_modules

    @property
    def document_private_members(self):
        return self._document_private_members

    @property
    def document_private_modules(self):
        return self._document_private_modules

    @property
    def initial_source_paths(self):
        return self._initial_source_paths

    @property
    def member_documenter_classes(self):
        return self._member_documenter_classes

    @property
    def module_documenter_class(self):
        return self._module_documenter_class

    @property
    def target_directory(self):
        return self._target_directory

    @property
    def root_documenter_class(self):
        return self._root_documenter_class



import uqbar.graphs
from uqbar.containers import UniqueTreeNode
from typing import Optional, Set


class Attachable(UniqueTreeNode):

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Mixins'

    ### INITIALIZER ###

    def __init__(self):
        self._edges: Set[uqbar.graphs.Edge] = set()

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        node = self._get_node()
        node_name = ''
        if node is not None:
            node_name = node._get_canonical_name()
        port_name = self._get_port_name()
        return '{}:{}'.format(node_name, port_name)

    def _get_node(self) -> Optional['uqbar.graphs.Node']:
        from uqbar.graphs.Node import Node
        parent = self.parent
        while parent is not None:
            if isinstance(parent, Node):
                return parent
            parent = parent.parent
        return None

    def _get_port_name(self) -> str:
        graph_order = self.graph_order
        node = self._get_node()
        if node is not None:
            node_graph_order = node.graph_order
            graph_order = graph_order[len(node_graph_order):]
        else:
            graph_order = (0,)
        return 'f_' + '_'.join(str(x) for x in graph_order)

    ### PUBLIC METHODS ###

    def attach(
        self,
        target,
        is_directed=True,
        head_port_position=None,
        tail_port_position=None,
        **attributes
        ) -> 'uqbar.graphs.Edge':
        edge = uqbar.graphs.Edge(
            attributes=attributes,
            head_port_position=head_port_position,
            is_directed=is_directed,
            tail_port_position=tail_port_position,
            )
        edge.attach(self, target)
        return edge

    ### PUBLIC PROPERTIES ###

    @property
    def edges(self) -> Set['uqbar.graphs.Edge']:
        return set(self._edges)



import enum
from uqbar.strings import to_snake_case


def from_expr(cls, expr):
    if isinstance(expr, cls):
        return expr
    elif isinstance(expr, int):
        return cls(expr)
    elif isinstance(expr, str):
        coerced_expr = to_snake_case(expr.strip()).upper()
        try:
            return cls[coerced_expr]
        except KeyError:
            pass
        try:
            return cls[coerced_expr.title()]
        except KeyError:
            pass
        try:
            return cls[coerced_expr.replace('_', '')]
        except KeyError:
            pass
    elif expr is None:
        return cls(0)
    message = 'Cannot instantiate {} from {!r}.'.format(cls.__name__, expr)
    raise ValueError(message)


class IntEnumeration(enum.IntEnum):
    """
    Enumeration which behaves like an integer.

    ::

        >>> import uqbar.enums
        >>> class MyEnum(uqbar.enums.IntEnumeration):
        ...     FOO = -1
        ...     BAR = 0
        ...     BAZ = 1
        ...     QUUX = 2

    ::

        >>> MyEnum.FOO < MyEnum.BAR
        True

    Does compare to similarly-valued objects:

    ::

        >>> MyEnum.BAR == False
        True

    """
    pass

    # ### SPECIAL METHODS ### #

    def __dir__(self):
        names = [
            '__class__',
            '__doc__',
            '__format__',
            '__members__',
            '__module__',
            '__repr__',
            'from_expr',
            ]
        names += self._member_names_
        names += [
            ]
        return sorted(names)

    def __repr__(self):
        return '{}.{}'.format(
            type(self).__name__,
            self.name,
            )

    # ### PUBLIC METHODS ### #

    @classmethod
    def from_expr(cls, expr):
        r'''Convenience constructor for enumerations.

        Returns new enumeration item.
        '''
        return from_expr(cls, expr)


class StrictEnumeration(enum.Enum):
    """
    Sortable enumeration which does not compare to objects not of its type.

    ::

        >>> import uqbar.enums
        >>> class MyEnum(uqbar.enums.StrictEnumeration):
        ...     FOO = -1
        ...     BAR = 0
        ...     BAZ = 1
        ...     QUUX = 2

    ::

        >>> MyEnum.FOO < MyEnum.BAR
        True

    Does not compare to similarly-valued objects:

    ::

        >>> MyEnum.BAR == False
        False

    """

    # ### SPECIAL METHODS ### #

    def __dir__(self):
        names = [
            '__class__',
            '__doc__',
            '__format__',
            '__members__',
            '__module__',
            '__repr__',
            'from_expr',
            ]
        names += self._member_names_
        names += [
            ]
        return sorted(names)

    def __float__(self):
        return float(self.value)

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __int__(self):
        return int(self.value)

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __repr__(self):
        return '{}.{}'.format(
            type(self).__name__,
            self.name,
            )

    # ### PUBLIC METHODS ### #

    @classmethod
    def from_expr(cls, expr):
        r'''Convenience constructor for enumerations.

        Returns new enumeration item.
        '''
        return from_expr(cls, expr)



"""
Tools for building Graphviz graphs.
"""

from .Attachable import Attachable  # noqa
from .Attributes import Attributes  # noqa
from .Edge import Edge  # noqa
from .Graph import Graph  # noqa
from .Node import Node  # noqa
from .RecordField import RecordField  # noqa
from .RecordGroup import RecordGroup  # noqa
from .HRule import HRule  # noqa
from .LineBreak import LineBreak  # noqa
from .Table import Table  # noqa
from .TableCell import TableCell  # noqa
from .TableRow import TableRow  # noqa
from .Text import Text  # noqa
from .VRule import VRule  # noqa



from typing import Mapping, Tuple, Union
from uqbar.graphs.Attributes import Attributes
from uqbar.containers import UniqueTreeContainer
from uqbar.graphs import Attachable


class TableCell(UniqueTreeContainer, Attachable):
    """
    A Graphviz HTML table.

    ::

        >>> import uqbar.graphs
        >>> table_cell = uqbar.graphs.TableCell()
        >>> print(format(table_cell, 'graphviz'))
        <TD></TD>

    ::

        >>> table_cell = uqbar.graphs.TableCell(
        ...     attributes={
        ...         'border': 5,
        ...         'bgcolor': 'blue',
        ...         },
        ...     )
        >>> print(format(table_cell, 'graphviz'))
        <TD BGCOLOR="blue" BORDER="5"></TD>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'HTML Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Union[Mapping[str, object], Attributes]=None,
        name: str=None
        ) -> None:
        from uqbar.graphs import Text
        if isinstance(children, str):
            children = [Text(children)]
        UniqueTreeContainer.__init__(
            self,
            children=children,
            name=name,
            )
        Attachable.__init__(self)
        self._attributes = Attributes('table_cell', **(attributes or {}))

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        start, stop = '<TD', '</TD>'
        if self.edges:
            start += ' PORT="{}"'.format(self._get_port_name())
        attributes = format(self._attributes, 'html')
        if attributes:
            start += ' {}'.format(attributes)
        start += '>'
        result.append(start)
        for child in self:
            result.append(format(child, 'graphviz'))
        result.append(stop)
        return ''.join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.Table,
            uqbar.graphs.LineBreak,
            uqbar.graphs.Text,
            )



"""
Uqbar Sphinx styling extension.

- Adds additional information to autodoc entries.
- Adds additional CSS.

Install by adding ``'uqbar.sphinx.style'`` to the ``extensions`` list in your
Sphinx configuration.
"""
import importlib
import inspect
import pathlib
import sphinx  # type:ignore
import uqbar.io
from docutils import nodes
from sphinx import addnodes  # type: ignore
from typing import Dict


def handle_class(signature_node, module, object_name, cache):
    """
    Styles ``autoclass`` entries.

    Adds ``abstract`` prefix to abstract classes.
    """
    class_ = getattr(module, object_name, None)
    if class_ is None:
        return
    if class_ not in cache:
        cache[class_] = {}
        attributes = inspect.classify_class_attrs(class_)
        for attribute in attributes:
            cache[class_][attribute.name] = attribute
    if inspect.isabstract(class_):
        emphasis = nodes.emphasis(
            'abstract ',
            'abstract ',
            classes=['property'],
            )
        signature_node.insert(0, emphasis)


def handle_method(signature_node, module, object_name, cache):
    """
    Styles ``automethod`` entries.

    Adds ``abstract`` prefix to abstract methods.

    Adds link to originating class for inherited methods.
    """
    cls_name, attr_name = object_name.split('.')
    class_ = getattr(module, cls_name, None)
    if class_ is None:
        return
    attr = getattr(class_, attr_name)
    inspected_attr = cache[class_][attr_name]
    defining_class = inspected_attr.defining_class
    if defining_class is not class_:
        reftarget = '{}.{}'.format(
            defining_class.__module__,
            defining_class.__name__,
            )
        xref_node = addnodes.pending_xref(
            '',
            refdomain='py',
            refexplicit=True,
            reftype='class',
            reftarget=reftarget,
            )
        name_node = nodes.literal(
            '',
            '{}'.format(defining_class.__name__),
            classes=['descclassname'],
            )
        xref_node.append(name_node)
        desc_annotation = list(signature_node.traverse(
            addnodes.desc_annotation))
        index = len(desc_annotation)
        class_annotation = addnodes.desc_addname()
        class_annotation.extend([nodes.Text('('), xref_node, nodes.Text(').')])
        class_annotation['xml:space'] = 'preserve'
        signature_node.insert(index, class_annotation)
    if getattr(attr, '__isabstractmethod__', False):
        emphasis = nodes.emphasis(
            'abstract ',
            'abstract ',
            classes=['property'],
            )
        signature_node.insert(0, emphasis)


def on_doctree_read(
    app: sphinx.application.Sphinx,
    document,
    ):
    """
    Hooks into Sphinx's ``doctree-read`` event.
    """
    cache: Dict[type, Dict[str, object]] = {}
    for desc_node in document.traverse(addnodes.desc):
        if desc_node.get('domain') != 'py':
            continue
        signature_node = desc_node.traverse(addnodes.desc_signature)[0]
        module_name = signature_node.get('module')
        object_name = signature_node.get('fullname')
        object_type = desc_node.get('objtype')
        module = importlib.import_module(module_name)
        if object_type == 'class':
            handle_class(signature_node, module, object_name, cache)
        elif object_type in (
            'method',
            'attribute',
            'staticmethod', 'classmethod'
            ):
            handle_method(signature_node, module, object_name, cache)


def on_builder_inited(app: sphinx.application.Sphinx):
    """
    Hooks into Sphinx's ``builder-inited`` event.

    Used for copying over CSS files to theme directory.
    """
    local_css_path = pathlib.Path(__file__).parent / 'uqbar.css'
    theme_css_path = (
        pathlib.Path(app.srcdir) /
        app.config.html_static_path[0] /
        'uqbar.css'
        )
    with local_css_path.open('r') as file_pointer:
        local_css_contents = file_pointer.read()
    uqbar.io.write(local_css_contents, theme_css_path)


def setup(app: sphinx.application.Sphinx):
    """
    Sets up Sphinx extension.
    """
    app.connect('doctree-read', on_doctree_read)
    app.connect('builder-inited', on_builder_inited)
    app.add_stylesheet('uqbar.css')



from typing import List, MutableMapping, Tuple  # noqa
from uqbar.apis.MemberDocumenter import MemberDocumenter
from uqbar.apis.ModuleDocumenter import ModuleDocumenter


class SummarizingModuleDocumenter(ModuleDocumenter):
    """
    A summarizing module documenter.

    Organizes member documenters by their *documentation section*.

    Treats *nominative* submodule documenters (almost) as if they were locally
    defined members. This means that a documenter for a package which contains
    a module which contains *only* one class or function named identically to
    that module will treat that class or function as though it was defined in
    the package's ``__init__.py``. These nominative submodule members are
    organized in each documentation section via an *autosummary* table, rather
    than including their documentation directly.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.SummarizingModuleDocumenter(
        ...     'uqbar.io',
        ...     module_documenters=[
        ...         uqbar.apis.ModuleDocumenter('uqbar.io.Timer'),
        ...         ],
        ...     )
        >>> print(str(documenter))
        .. _uqbar--io:
        <BLANKLINE>
        io
        ==
        <BLANKLINE>
        .. automodule:: uqbar.io
        <BLANKLINE>
        .. currentmodule:: uqbar.io
        <BLANKLINE>
        .. container:: svg-container
        <BLANKLINE>
           .. inheritance-diagram:: uqbar
              :lineage: uqbar.io
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Classes
           :class: section-header
        <BLANKLINE>
        .. toctree::
           :hidden:
        <BLANKLINE>
           Timer
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~Timer.Timer
        <BLANKLINE>
        .. raw:: html
        <BLANKLINE>
           <hr/>
        <BLANKLINE>
        .. rubric:: Functions
           :class: section-header
        <BLANKLINE>
        .. autosummary::
           :nosignatures:
        <BLANKLINE>
           ~find_common_prefix
           ~find_executable
           ~relative_to
           ~walk
           ~write
        <BLANKLINE>
        .. autofunction:: find_common_prefix
        <BLANKLINE>
        .. autofunction:: find_executable
        <BLANKLINE>
        .. autofunction:: relative_to
        <BLANKLINE>
        .. autofunction:: walk
        <BLANKLINE>
        .. autofunction:: write

    :param package_path: the module path of the module to document
    :param document_private_members: whether to documenter private module members
    :param member_documenter_classes: a list of
        :py:class:`~uqbar.apis.MemberDocumenter` subclasses, defining what classes
        to use to identify and document module members
    :param module_documenters: a list of of documenters for submodules and
        subpackages of the documented module; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    :param api_builder: an :py:class:`~uqbar.apis.APIBuilder` instance
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Documenters'

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        result = self._build_preamble()
        package_path = self.package_path.partition('.')[0]
        lineage_path = self.package_path
        result.extend([
            '',
            '.. container:: svg-container',
            '',
            '   .. inheritance-diagram:: {}'.format(package_path),
            '      :lineage: {}'.format(lineage_path),
            ])
        if self.is_nominative:
            result.extend(['', str(self.member_documenters[0])])
        else:
            if self.is_package:
                subpackage_documenters = [
                    _ for _ in self.module_documenters or []
                    if _.is_package or not _.is_nominative
                    ]
                if subpackage_documenters:
                    result.extend([
                        '',
                        '.. raw:: html',
                        '',
                        '   <hr/>',
                        '',
                        '.. rubric:: Subpackages',
                        '   :class: section-header',
                        ])
                    result.extend(self._build_toc(
                        subpackage_documenters,
                        show_full_paths=True,
                        ))
            for section, documenters in self.member_documenters_by_section:
                result.extend([
                    '',
                    '.. raw:: html',
                    '',
                    '   <hr/>',
                    '',
                    '.. rubric:: {}'.format(section),
                    '   :class: section-header',
                    ])
                local_documenters = [
                    documenter for documenter in documenters
                    if documenter.client.__module__ == self.package_path
                    ]
                result.extend(self._build_toc(documenters))
                for local_documenter in local_documenters:
                    result.extend(['', str(local_documenter)])
        return '\n'.join(result)

    ### PRIVATE METHODS ###

    def _build_toc(
        self,
        documenters,
        show_full_paths: bool=False,
        **kwargs
    ) -> List[str]:
        result: List[str] = []
        if not documenters:
            return result
        toctree_paths = set()
        for documenter in documenters:
            path = documenter.package_path.partition(
                self.package_path)[-1]
            if not isinstance(documenter, ModuleDocumenter):
                path = path.rpartition('.')[0]
            elif documenter.is_package:
                path += '/index'
            if path.startswith('.'):
                path = path[1:]
            if path:
                toctree_paths.add(path)
        if toctree_paths:
            result.extend(['', '.. toctree::', '   :hidden:', ''])
            for toctree_path in sorted(toctree_paths):
                result.append('   {}'.format(toctree_path))
        result.extend([
            '',
            '.. autosummary::',
            '   :nosignatures:',
            '',
            ])
        for documenter in documenters:
            template = '   ~{}'
            if show_full_paths:
                template = '   {}'
            path = documenter.package_path.rpartition(
                self.package_path + '.')[-1]
            result.append(template.format(path))
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def member_documenters_by_section(self) -> List[Tuple[str, List[MemberDocumenter]]]:
        result: MutableMapping[str, List[MemberDocumenter]] = {}
        for documenter in self.member_documenters:
            result.setdefault(
                documenter.documentation_section, []).append(documenter)
        for module_documenter in self.module_documenters or []:
            if not module_documenter.is_nominative:
                continue
            documenter = module_documenter.member_documenters[0]
            result.setdefault(
                documenter.documentation_section, []).append(documenter)
        return sorted(result.items())



import enum
import inspect
from typing import cast
from uqbar.apis.MemberDocumenter import MemberDocumenter


class ClassDocumenter(MemberDocumenter):
    """
    A basic class documenter.

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.apis.ClassDocumenter.ClassDocumenter'
        >>> documenter = uqbar.apis.ClassDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autoclass:: ClassDocumenter
           :members:
           :undoc-members:

    .. tip::

       Subclass :py:class:`~uqbar.apis.ClassDocumenter` to implement your own
       custom class documentation output. You'll need to provide your desired
       reStructuredText output via an overridden
       :py:meth:`~uqbar.apis.ClassDocumenter.ClassDocumenter.__str__`
       implementation.

       See :py:class:`~uqbar.apis.SummarizingClassDocumenter` for an example.

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Documenters'

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        name = getattr(self.client, '__name__')
        if issubclass(self.client, Exception):  # type: ignore
            return '.. autoexception:: {}'.format(name)
        return '\n'.join([
            '.. autoclass:: {}'.format(name),
            '   :members:',
            '   :undoc-members:',
            ])

    ### PUBLIC METHODS ###

    @classmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        return (
            isinstance(client, type) and
            getattr(client, '__module__') == module_path
            )

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_section(self) -> str:
        client = cast(type, self.client)
        if hasattr(client, '__documentation_section__'):
            section = getattr(client, '__documentation_section__')
            if section is not None:
                return section
        if inspect.isabstract(client):
            return 'Abstract Classes'
        elif issubclass(client, enum.Enum):
            return 'Enumerations'
        elif issubclass(client, Exception):
            return 'Exceptions'
        return 'Classes'



import copy
from uqbar.containers.UniqueTreeNode import UniqueTreeNode


class UniqueTreeContainer(UniqueTreeNode):
    """
    A container node in a "unique" tree.

    Container nodes may contain zero or more other nodes.

    Unique tree nodes may have at most one parent and may appear only once in
    the tree.
    """

    ### INITIALIZER ###

    def __init__(self, children=None, name=None):
        UniqueTreeNode.__init__(self, name=name)
        self._children = []
        self._named_children = {}
        if children is not None:
            self[:] = children

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        if isinstance(expr, str):
            return expr in self._named_children
        for x in self._children:
            if x is expr:
                return True
        return False

    def __delitem__(self, i):
        if isinstance(i, str):
            children = tuple(self._named_children[i])
            for child in children:
                parent = child.parent
                del(parent[parent.index(child)])
            return
        if isinstance(i, int):
            if i < 0:
                i = len(self) + i
            i = slice(i, i + 1)
        self.__setitem__(i, [])
        self._mark_entire_tree_for_later_update()

    def __getitem__(self, expr):
        if isinstance(expr, (int, slice)):
            return self._children[expr]
        elif isinstance(expr, str):
            result = sorted(
                self._named_children[expr],
                key=lambda x: x.graph_order,
                )
            if len(result) == 1:
                return result[0]
            return result
        raise ValueError(expr)

    def __iter__(self):
        for child in self._children:
            yield child

    def __len__(self):
        return len(self._children)

    def __setitem__(self, i, expr):
        if isinstance(i, int):
            expr = self._prepare_setitem_single(expr)
            assert isinstance(expr, self._node_class)
            old = self[i]
            if expr in self.parentage:
                raise ValueError('Cannot set parent node as child.')
            old._set_parent(None)
            expr._set_parent(self)
            self._children.insert(i, expr)
        else:
            expr = self._prepare_setitem_multiple(expr)
            if isinstance(expr, UniqueTreeContainer):
                # Prevent mutating while iterating by copying.
                expr = expr[:]
            assert all(isinstance(x, self._node_class) for x in expr)
            if i.start == i.stop and i.start is not None \
                and i.stop is not None and i.start <= -len(self):
                start, stop = 0, 0
            else:
                start, stop, stride = i.indices(len(self))
            old = self[start:stop]
            parentage = self.parentage
            if any(node in parentage for node in expr):
                raise ValueError('Cannot set parent node as child.')
            for node in old:
                node._set_parent(None)
            for node in expr:
                node._set_parent(self)
            self._children.__setitem__(slice(start, start), expr)
        self._mark_entire_tree_for_later_update()

    ### PRIVATE METHODS ###

    def _cache_named_children(self):
        name_dictionary = super(UniqueTreeContainer, self)._cache_named_children()
        if hasattr(self, '_named_children'):
            for name, children in self._named_children.items():
                name_dictionary[name] = copy.copy(children)
        return name_dictionary

    def _prepare_setitem_single(self, expr):
        return expr

    def _prepare_setitem_multiple(self, expr):
        return expr

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self):
        return UniqueTreeNode

    ### PUBLIC METHODS ###

    def append(self, expr):
        self.__setitem__(
            slice(len(self), len(self)),
            [expr]
            )

    def depth_first(self, top_down=True):
        for child in tuple(self):
            if top_down:
                yield child
            if isinstance(child, UniqueTreeContainer):
                yield from child.depth_first(top_down=top_down)
            if not top_down:
                yield child

    def extend(self, expr):
        self.__setitem__(
            slice(len(self), len(self)),
            expr
            )

    def index(self, expr):
        for i, child in enumerate(self._children):
            if child is expr:
                return i
        else:
            message = '{!r} not in {!r}.'
            message = message.format(expr, self)
            raise ValueError(message)

    def insert(self, i, expr):
        self.__setitem__(
            slice(i, i),
            [expr]
            )

    def pop(self, i=-1):
        node = self[i]
        del(self[i])
        return node

    def recurse(self):
        for child in self:
            yield child
            if isinstance(child, type(self)):
                for grandchild in child.recurse():
                    yield grandchild

    def remove(self, node):
        i = self.index(node)
        del(self[i])

    ### PUBLIC PROPERTIES ###

    @property
    def children(self):
        return tuple(self._children)



"""
Tools for auto-generating API documention.
"""
import importlib
import pathlib

from .APIBuilder import APIBuilder
from .ClassDocumenter import ClassDocumenter
from .FunctionDocumenter import FunctionDocumenter
from .InheritanceGraph import InheritanceGraph
from .MemberDocumenter import MemberDocumenter
from .ModuleDocumenter import ModuleDocumenter
from .ModuleNode import ModuleNode
from .PackageNode import PackageNode
from .RootDocumenter import RootDocumenter
from .SummarizingClassDocumenter import SummarizingClassDocumenter
from .SummarizingModuleDocumenter import SummarizingModuleDocumenter
from .SummarizingRootDocumenter import SummarizingRootDocumenter


def collect_source_paths(source_paths, recurse_subpackages=True):
    visited_paths = set()
    path_stack = []
    for source_path in source_paths:
        source_path = pathlib.Path(source_path).resolve().absolute()
        assert source_path.exists()
        path_stack.append(source_path)
    while path_stack:
        current_path = path_stack.pop()
        if current_path in visited_paths:
            continue
        if current_path.is_dir():
            if not (current_path / '__init__.py').exists():
                continue
            if recurse_subpackages:
                for path in current_path.iterdir():
                    if (path.is_dir() and (path / '__init__.py').exists()):
                        path_stack.append(path)
                    elif path.suffix in ('.py', '.pyx'):
                        visited_paths.add(path)
            else:
                visited_paths.add(current_path / '__init__.py')
                for path in current_path.iterdir():
                    if path.is_dir():
                        continue
                    # Permit nominative modules
                    if path.suffix in ('.py', '.pyx'):
                        package_path = source_path_to_package_path(path)
                        module = importlib.import_module(package_path)
                        if hasattr(module, path.with_suffix('').name):
                            visited_paths.add(path)
        elif current_path.suffix in ('.py', '.pyx'):
            visited_paths.add(current_path)
    return sorted(visited_paths)


def source_path_to_package_path(path):
    path = pathlib.Path(path)
    if not (path.parent / '__init__.py').exists():
        return path.with_suffix('').name
    root = path
    while (root.parent / '__init__.py').exists():
        root = root.parent
    path = path.with_suffix('')
    if path == root:
        return path.name
    parts = (root.name,) + path.relative_to(root).parts
    if parts[-1] == '__init__':
        parts = parts[:-1]
    return '.'.join(parts)



import types
from uqbar.apis.MemberDocumenter import MemberDocumenter


class FunctionDocumenter(MemberDocumenter):
    """
    A basic function documenter.

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.io.walk'
        >>> documenter = uqbar.apis.FunctionDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autofunction:: walk

    .. tip::

       Subclass :py:class:`~uqbar.apis.FunctionDocumenter` to
       implement your own custom module documentation output.
       You'll need to provide your desired reStructuredText output
       via an overridden
       :py:meth:`~uqbar.apis.FunctionDocumenter.FunctionDocumenter.__str__`
       implementation.

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Documenters'

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        return '.. autofunction:: {}'.format(
            getattr(self.client, '__name__'),
            )

    ### PUBLIC METHODS ###

    @classmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        return (
            isinstance(client, types.FunctionType) and
            client.__module__ == module_path
            )

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_section(self) -> str:
        return 'Functions'



from uqbar.containers import UniqueTreeContainer
from uqbar.graphs.Attributes import Attributes
from uqbar.graphs.Edge import Edge  # noqa
from uqbar.graphs.RecordField import RecordField
from uqbar.graphs.RecordGroup import RecordGroup
from uqbar.graphs.Table import Table
from typing import Iterable, Mapping, Set, Tuple, Union  # noqa


class Node(UniqueTreeContainer):
    """
    A Graphviz node.

    ::

        >>> import uqbar.graphs
        >>> node = uqbar.graphs.Node()
        >>> print(format(node, 'graphviz'))
        node_0;

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Core Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        children: Iterable[Union[RecordField, RecordGroup]]=None,
        *,
        attributes: Union[Mapping[str, object], Attributes]=None,
        name: str=None
        ) -> None:
        UniqueTreeContainer.__init__(self, name=name, children=children)
        self._attributes = Attributes('node', **(attributes or {}))
        self._edges: Set[Edge] = set()

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        node_definition = Attributes._format_value(
            self._get_canonical_name())
        result = [node_definition]
        attributes = self.attributes.copy()
        if len(self):
            if isinstance(self[0], Table):
                label = '<\n{}>'.format(format(self[0], 'graphviz'))
            else:
                label = ' | '.join(format(_, 'graphviz') for _ in self)
            attributes['label'] = label
        if len(attributes):
            attributes = format(attributes, 'graphviz').split('\n')
            result[0] = '{} {}'.format(result[0], attributes[0])
            result.extend(attributes[1:])
        else:
            result[-1] += ';'
        return '\n'.join(result)

    ### PRIVATE METHODS ###

    def _get_canonical_name(self) -> str:
        prefix = 'node'
        if self.name is not None:
            root = self.root
            if root:
                instances = root[self.name]
                if instances is not self:
                    return '{}_{}'.format(self.name, instances.index(self))
            return self.name
        elif self.graph_order:
            suffix = '_'.join(str(x) for x in self.graph_order)
        else:
            suffix = '0'
        return '{}_{}'.format(prefix, suffix)

    ### PUBLIC METHODS ###

    def attach(
        self,
        node,
        is_directed=True,
        head_port_position=None,
        tail_port_position=None,
        **attributes
        ) -> Edge:
        import uqbar.graphs
        edge = uqbar.graphs.Edge(
            attributes=attributes,
            head_port_position=head_port_position,
            is_directed=is_directed,
            tail_port_position=tail_port_position,
            )
        edge.attach(self, node)
        return edge

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.RecordField,
            uqbar.graphs.RecordGroup,
            uqbar.graphs.Table,
            )

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def edges(self) -> Set[Edge]:
        return set(self._edges)



import uqbar.containers


class ModuleNode(uqbar.containers.UniqueTreeNode):
    """
    A :py:class:`~uqbar.containers.UniqueTreeNode` subclass used during API
    construction to proxy modules.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Internals'

    ### INITIALIZER ###

    def __init__(
        self,
        name=None,
        documenter=None,
        source_path=None,
    ):
        super().__init__(name=name)
        self._documenter = documenter
        self._source_path = source_path

    ### SPECIAL METHODS ###

    def __str__(self):
        return '{}'.format(self.name)

    ### PUBLIC PROPERTIES ###

    @property
    def documenter(self):
        return self._documenter

    @documenter.setter
    def documenter(self, documenter):
        self._documenter = documenter

    @property
    def package_path(self):
        return self._name

    @property
    def source_path(self):
        return self._source_path

    @source_path.setter
    def source_path(self, source_path):
        self._source_path = source_path



import cProfile
import io
import pstats


class Profiler:
    """
    A context manager for profiling blocks of code.
    """

    def __enter__(self) -> 'Profiler':
        self._profiler = cProfile.Profile()
        self._profiler.enable()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._profiler.disable()
        stream = io.StringIO()
        profiler_stats = pstats.Stats(self._profiler, stream=stream)
        profiler_stats = profiler_stats.sort_stats('cumulative')
        profiler_stats.print_stats()
        print(stream.getvalue())
        profiler_stats.dump_stats('stats.profile')



import uqbar.graphs  # noqa
from typing import Mapping, Optional, Tuple, Union
from uqbar.graphs.Attributes import Attributes


class Edge(object):
    """
    A Graphviz edge.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Core Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        attributes: Union[Mapping[str, object], Attributes]=None,
        is_directed: bool=True,
        head_port_position: str=None,
        tail_port_position: str=None,
        ) -> None:
        self._attributes = Attributes('edge', **(attributes or {}))
        self._head: Optional[
            Union[uqbar.graphs.Node, uqbar.graphs.Attachable]] = None
        self._head_port_position = head_port_position
        self._is_directed = bool(is_directed)
        self._tail: Optional[
            Union[uqbar.graphs.Node, uqbar.graphs.Attachable]] = None
        self._tail_port_position = tail_port_position

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        from uqbar.graphs.Attachable import Attachable
        from uqbar.graphs.Node import Node
        connection = '->'
        if not self.is_directed:
            connection = '--'

        tail_parts = []
        tail_node = None
        tail_port_name = None
        if isinstance(self.tail, Node):
            tail_node = self.tail
        elif isinstance(self.tail, Attachable):
            tail_node = self.tail._get_node()
            tail_port_name = self.tail._get_port_name()
        if tail_node is None:
            raise ValueError
        tail_parts.append(tail_node._get_canonical_name())
        if tail_port_name:
            tail_parts.append(tail_port_name)
        if self.tail_port_position:
            tail_parts.append(self.tail_port_position)
        tail_name = ':'.join(
            Attributes._format_value(part)
            for part in tail_parts
            )

        head_parts = []
        head_node = None
        head_port_name = None
        if isinstance(self.head, Node):
            head_node = self.head
        elif isinstance(self.head, Attachable):
            head_node = self.head._get_node()
            head_port_name = self.head._get_port_name()
        if head_node is None:
            raise ValueError
        head_parts.append(head_node._get_canonical_name())
        if head_port_name:
            head_parts.append(head_port_name)
        if self.head_port_position:
            head_parts.append(self.head_port_position)
        head_name = ':'.join(
            Attributes._format_value(part)
            for part in head_parts
            )

        edge_definition = '{} {} {}'.format(tail_name, connection, head_name)
        result = [edge_definition]
        if len(self.attributes):
            attributes = format(self.attributes, 'graphviz').split('\n')
            result[0] = '{} {}'.format(result[0], attributes[0])
            result.extend(attributes[1:])
        else:
            result[-1] += ';'
        return '\n'.join(result)

    ### PRIVATE METHODS ###

    def _get_highest_parent(self) -> 'uqbar.graphs.Graph':
        if self.tail is None:
            raise ValueError(self.tail)
        elif self.head is None:
            raise ValueError(self.head)
        highest_parent: Optional[uqbar.graphs.Graph] = None
        tail_parentage = list(self.tail.parentage[1:])
        head_parentage = list(self.head.parentage[1:])
        while (
            len(tail_parentage) and
            len(head_parentage) and
            tail_parentage[-1] is head_parentage[-1]
            ):
            highest_parent = tail_parentage[-1]
            tail_parentage.pop()
            head_parentage.pop()
        if highest_parent is None:
            message = 'highest parent can not be none.'
            raise Exception(message)
        return highest_parent

    ### PUBLIC METHODS ###

    def attach(
        self,
        tail: Union['uqbar.graphs.Node', 'uqbar.graphs.Attachable'],
        head: Union['uqbar.graphs.Node', 'uqbar.graphs.Attachable'],
        ) -> 'Edge':
        from uqbar.graphs import Node
        from uqbar.graphs import Attachable
        prototype = (Node, Attachable)
        assert isinstance(tail, prototype)
        assert isinstance(head, prototype)
        self.detach()
        tail._edges.add(self)
        head._edges.add(self)
        self._tail = tail
        self._head = head
        return self

    def detach(self) -> 'Edge':
        if self.tail is not None:
            self.tail._edges.remove(self)
            self._tail = None
        if self.head is not None:
            self.head._edges.remove(self)
            self._head = None
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def head(self) -> Optional[Union['uqbar.graphs.Node', 'uqbar.graphs.Attachable']]:
        return self._head

    @property
    def head_graph_order(self) -> Tuple[int, ...]:
        if self.head is None:
            return ()
        return self.head.graph_order

    @property
    def head_port_position(self) -> Optional[str]:
        return self._head_port_position

    @property
    def is_directed(self) -> bool:
        return self._is_directed

    @property
    def tail(self) -> Optional[Union['uqbar.graphs.Node', 'uqbar.graphs.Attachable']]:
        return self._tail

    @property
    def tail_graph_order(self) -> Tuple[int, ...]:
        if self.tail is None:
            return ()
        return self.tail.graph_order

    @property
    def tail_port_position(self) -> Optional[str]:
        return self._tail_port_position



import collections
import copy
import typing


class UniqueTreeNode:
    """
    A node in a "unique" tree.

    Unique tree nodes may have at most one parent and may appear only once in
    the tree.
    """

    ### CLASS VARIABLES ###

    _state_flag_names: typing.Tuple[str, ...] = ()

    ### INITIALIZER ###

    def __init__(self, name: str=None) -> None:
        self._name = name
        self._parent = None

    ### PRIVATE METHODS ###

    def _cache_named_children(self):
        name_dictionary = {}
        if self.name is not None:
            if self.name not in name_dictionary:
                name_dictionary[self.name] = set()
            name_dictionary[self.name].add(self)
        return name_dictionary

    @classmethod
    def _iterate_nodes(cls, nodes):
        for x in nodes:
            yield x
            if hasattr(x, '_children'):
                for y in cls._iterate_nodes(x):
                    yield y

    def _get_node_state_flags(self):
        state_flags = {}
        for name in self._state_flag_names:
            state_flags[name] = True
            for node in self.parentage:
                if not getattr(node, name):
                    state_flags[name] = False
                    break
        return state_flags

    def _mark_entire_tree_for_later_update(self):
        for node in self.parentage:
            for name in self._state_flag_names:
                setattr(node, name, False)

    def _remove_from_parent(self):
        if self._parent is not None:
            if self in self._parent:
                self._parent._children.remove(self)
        self._parent = None

    def _remove_named_children_from_parentage(self, name_dictionary):
        if self._parent is not None and name_dictionary:
            for parent in self.parentage[1:]:
                named_children = parent._named_children
                for name in name_dictionary:
                    for node in name_dictionary[name]:
                        named_children[name].remove(node)
                    if not named_children[name]:
                        del(named_children[name])

    def _restore_named_children_to_parentage(self, name_dictionary):
        if self._parent is not None and name_dictionary:
            for parent in self.parentage[1:]:
                named_children = parent._named_children
                for name in name_dictionary:
                    if name in named_children:
                        named_children[name].update(name_dictionary[name])
                    else:
                        named_children[name] = copy.copy(name_dictionary[name])

    def _set_parent(self, new_parent):
        named_children = self._cache_named_children()
        self._remove_named_children_from_parentage(named_children)
        self._remove_from_parent()
        self._parent = new_parent
        self._restore_named_children_to_parentage(named_children)
        self._mark_entire_tree_for_later_update()

    ### PUBLIC METHODS ###

    def precede_by(self, expr):
        if not isinstance(expr, collections.Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index:index] = expr

    def replace_with(self, expr):
        if not isinstance(expr, collections.Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index:index + 1] = expr

    def succeed_by(self, expr):
        if not isinstance(expr, collections.Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index + 1:index + 1] = expr

    ### PUBLIC PROPERTIES ###

    @property
    def depth(self):
        return len(self.parentage) - 1

    @property
    def graph_order(self):
        parentage = tuple(reversed(self.parentage))
        graph_order = []
        for i in range(len(parentage) - 1):
            parent, child = parentage[i:i + 2]
            graph_order.append(parent.index(child))
        return tuple(graph_order)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, expr):
        assert isinstance(expr, (str, type(None)))
        old_name = self._name
        for parent in self.parentage[1:]:
            named_children = parent._named_children
            if old_name is not None:
                named_children[old_name].remove(self)
                if not named_children[old_name]:
                    del named_children[old_name]
            if expr is not None:
                if expr not in named_children:
                    named_children[expr] = set([self])
                else:
                    named_children[expr].add(self)
        self._name = expr

    @property
    def parent(self):
        return self._parent

    @property
    def parentage(self):
        parentage = []
        node = self
        while node is not None:
            parentage.append(node)
            node = node.parent
        return tuple(parentage)

    @property
    def root(self):
        proper_parentage = self.parentage[1:]
        if not proper_parentage:
            return None
        return proper_parentage[-1]



from uqbar.cli.CLI import CLI


class CLIAggregator(CLI):
    '''
    Aggregates CLI scripts.

    ::

        >>> import uqbar.cli
        >>> script = uqbar.cli.CLIAggregator()
        >>> try:
        ...     script('--help')
        ... except SystemExit:
        ...     pass
        ...
        usage: cliaggregator [-h] [--version] {help,list} ...
        <BLANKLINE>
        optional arguments:
          -h, --help   show this help message and exit
          --version    show program's version number and exit
        <BLANKLINE>
        subcommands:
          {help,list}
            help       print subcommand help
            list       list subcommands

    '''

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __call__(self, arguments=None):
        if arguments is None:
            arguments = self.argument_parser.parse_known_args()
        else:
            if isinstance(arguments, str):
                arguments = arguments.split()
            elif not isinstance(arguments, (list, tuple)):
                message = 'must be str, list, tuple or none: {!r}.'
                message = message.format(arguments)
                raise ValueError(message)
            arguments = self.argument_parser.parse_known_args(arguments)
        self._process_args(arguments)

    ### PRIVATE METHODS ###

    def _handle_help_command(self, unknown_args):
        aliases = self.cli_aliases
        program_names = self.cli_program_names
        cli_class = None
        if len(unknown_args) == 2 and \
            unknown_args[0] in aliases and \
            unknown_args[1] in aliases[unknown_args[0]]:
            cli_class = \
                aliases[unknown_args[0]][unknown_args[1]]
        elif len(unknown_args) == 1 and \
            unknown_args[0] in aliases and \
            not isinstance(aliases[unknown_args[0]], dict):
            cli_class = aliases[unknown_args[0]]
        elif len(unknown_args) == 1 and \
            unknown_args[0] in program_names:
            cli_class = program_names[unknown_args[0]]
        elif not len(unknown_args):
            self(['--help'])
            return
        if cli_class:
            instance = cli_class()
            print(instance.formatted_help)
        else:
            print('Cannot resolve {} to subcommand.'.format(unknown_args))

    def _handle_list_command(self):
        by_scripting_group = {}
        for cli_class in self.cli_classes:
            instance = cli_class()
            scripting_group = getattr(instance, 'scripting_group', None)
            group = by_scripting_group.setdefault(scripting_group, [])
            group.append(instance)
        print()
        if None in by_scripting_group:
            group = by_scripting_group.pop(None)
            for instance in sorted(group, key=lambda x: x.alias):
                message = '{}: {}'.format(
                    instance.alias,
                    instance.short_description,
                    )
                print(message)
            print()
        for group, instances in sorted(by_scripting_group.items()):
            print('[{}]'.format(group))
            for instance in sorted(instances, key=lambda x: x.alias):
                message = '    {}: {}'.format(
                    instance.alias,
                    instance.short_description,
                    )
                print(message)
            print()

    def _process_args(self, arguments):
        arguments, unknown_args = arguments
        if arguments.subparser_name == 'help':
            self._handle_help_command(unknown_args)
        elif arguments.subparser_name == 'list':
            self._handle_list_command()
        else:
            if hasattr(arguments, 'subsubparser_name'):
                cli_class = self.cli_aliases[
                    arguments.subparser_name][arguments.subsubparser_name]
            elif getattr(arguments, 'subparser_name'):
                cli_class = \
                    self.cli_aliases[arguments.subparser_name]
            elif getattr(arguments, 'subparser_name') is None:
                self(['--help'])
                return
            instance = cli_class()
            instance(unknown_args)

    def _setup_argument_parser(self, parser):
        subparsers = parser.add_subparsers(
            dest='subparser_name',
            title='subcommands',
            )
        subparsers.add_parser(
            'help',
            add_help=False,
            help='print subcommand help'
            )
        subparsers.add_parser(
            'list',
            add_help=False,
            help='list subcommands',
            )
        alias_map = self.cli_aliases
        for key in sorted(alias_map):
            if not isinstance(alias_map[key], dict):
                cli_class = alias_map[key]
                instance = cli_class()
                subparsers.add_parser(
                    key,
                    add_help=False,
                    help=instance.short_description,
                    )
            else:
                subkeys = sorted(alias_map[key])
                group_subparser = subparsers.add_parser(
                    key,
                    help='{{{}}} subcommand(s)'.format(', '.join(subkeys)),
                    )
                group_subparsers = group_subparser.add_subparsers(
                    dest='subsubparser_name'.format(key),
                    title='{} subcommands'.format(key),
                    )
                for subkey in subkeys:
                    cli_class = alias_map[key][subkey]
                    instance = cli_class()
                    group_subparsers.add_parser(
                        subkey,
                        add_help=False,
                        help=instance.short_description
                        )

    ### PUBLIC PROPERTIES ###

    @property
    def cli_aliases(self):
        r'''Developer script aliases.
        '''
        scripting_groups = []
        aliases = {}
        for cli_class in self.cli_classes:
            instance = cli_class()
            if getattr(instance, 'alias', None):
                scripting_group = getattr(instance, 'scripting_group', None)
                if scripting_group:
                    scripting_groups.append(scripting_group)
                    entry = (scripting_group, instance.alias)
                    if (scripting_group,) in aliases:
                        message = 'alias conflict between scripting group'
                        message += ' {!r} and {}'
                        message = message.format(
                            scripting_group,
                            aliases[(scripting_group,)].__name__,
                            )
                        raise Exception(message)
                    if entry in aliases:
                        message = 'alias conflict between {} and {}'
                        message = message.format(
                            aliases[entry].__name__,
                            cli_class.__name__,
                            )
                        raise Exception(message)
                    aliases[entry] = cli_class
                else:
                    entry = (instance.alias,)
                    if entry in scripting_groups:
                        message = 'alias conflict between {}'
                        message += ' and scripting group {!r}'
                        message = message.format(
                            cli_class.__name__,
                            instance.alias,
                            )
                        raise Exception(message)
                    if entry in aliases:
                        message = 'alias conflict be {} and {}'
                        message = message.format(
                            cli_class.__name__,
                            aliases[entry],
                            )
                        raise Exception(message)
                    aliases[(instance.alias,)] = cli_class
            else:
                if instance.program_name in scripting_groups:
                    message = 'Alias conflict between {}'
                    message += ' and scripting group {!r}'
                    message = message.format(
                        cli_class.__name__,
                        instance.program_name,
                        )
                    raise Exception(message)
                aliases[(instance.program_name,)] = cli_class
        alias_map = {}
        for key, value in aliases.items():
            if len(key) == 1:
                alias_map[key[0]] = value
            else:
                if key[0] not in alias_map:
                    alias_map[key[0]] = {}
                alias_map[key[0]][key[1]] = value
        return alias_map

    @property
    def cli_classes(self):
        r'''Developer scripts classes.
        '''
        return []

    @property
    def cli_program_names(self):
        r'''Developer script program names.
        '''
        program_names = {}
        for cli_class in self.cli_classes:
            instance = cli_class()
            program_names[instance.program_name] = cli_class
        return program_names



import time
import typing


class Timer:
    """
    A context manager for timing blocks of code.

    This context manager is not reentrant. Use a separate instance when nesting
    multiple timers.

    :param exit_message: message to print on entering the context
    :param enter_message: message to print on exiting the context
    :param verbose: whether to print output
    """

    def __init__(
        self,
        exit_message: str=None,
        enter_message: str=None,
        verbose: bool=True,
        ) -> None:
        if enter_message is not None:
            enter_message = str(enter_message)
        self._enter_message = enter_message
        if exit_message is not None:
            exit_message = str(exit_message)
        self._exit_message = exit_message
        self._start_time: typing.Optional[float] = None
        self._stop_time: typing.Optional[float] = None
        self._verbose = bool(verbose)

    ### SPECIAL METHODS ###

    def __enter__(self) -> 'Timer':
        if self.enter_message and self.verbose:
            print(self.enter_message)
        self._stop_time = None
        self._start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._stop_time = time.time()
        if self.exit_message and self.verbose:
            print(self.exit_message, self.elapsed_time)

    ### PUBLIC PROPERTIES ###

    @property
    def elapsed_time(self) -> typing.Union[float, None]:
        if self.start_time is not None:
            if self.stop_time is not None:
                return self.stop_time - self.start_time
            return time.time() - self.start_time
        return None

    @property
    def enter_message(self) -> typing.Union[str, None]:
        return self._enter_message

    @property
    def exit_message(self) -> typing.Union[str, None]:
        return self._exit_message

    @property
    def start_time(self) -> typing.Union[float, None]:
        return self._start_time

    @property
    def stop_time(self) -> typing.Union[float, None]:
        return self._stop_time

    @property
    def verbose(self) -> bool:
        return self._verbose



from typing import Tuple
from uqbar.containers import UniqueTreeContainer


class TableRow(UniqueTreeContainer):
    """
    A Graphviz HTML table row.

    ::

        >>> import uqbar.graphs
        >>> table_row = uqbar.graphs.TableRow()
        >>> print(format(table_row, 'graphviz'))
        <TR></TR>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'HTML Classes'

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        result.append('<TR>')
        for child in self:
            for line in format(child, 'graphviz').splitlines():
                result.append('    {}'.format(line))
        result.append('</TR>')
        join_character = ''
        if len(result) > 2:
            join_character = '\n'
        return join_character.join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.TableCell,
            uqbar.graphs.VRule,
            )



import uqbar.graphs  # noqa
from uqbar.containers import UniqueTreeNode
from typing import Optional
from uqbar.graphs import Attachable


class RecordField(Attachable, UniqueTreeNode):
    """
    A Graphviz record field.

    ::

        >>> import uqbar.graphs
        >>> field = uqbar.graphs.RecordField(label='My Label')
        >>> print(format(field, 'graphviz'))
        <f_0> My Label

    Port names are generated based on the graph order of the field within its
    parent node:

    ::

        >>> other_field = uqbar.graphs.RecordField(label='Other Label')
        >>> group = uqbar.graphs.RecordGroup()
        >>> node = uqbar.graphs.Node(attributes=dict(shape='Mrecord'))
        >>> cluster = uqbar.graphs.Graph(is_cluster=True)
        >>> graph = uqbar.graphs.Graph()

    ::

        >>> graph.append(cluster)
        >>> cluster.append(node)
        >>> node.append(group)
        >>> group.extend([other_field, field])
        >>> print(format(field, 'graphviz'))
        <f_0_1> My Label

    If a node contains record fields or groups, their format contributions
    override any string label in the node's attributes:

    ::

        >>> print(format(node, 'graphviz'))
        node_0_0 [label="{ <f_0_0> Other Label | <f_0_1> My Label }",
            shape=Mrecord];

    ::

        >>> node.attributes['label'] = 'Foo'
        >>> print(format(node, 'graphviz'))
        node_0_0 [label="{ <f_0_0> Other Label | <f_0_1> My Label }",
            shape=Mrecord];

    ::

        >>> node[:] = []
        >>> print(format(node, 'graphviz'))
        node_0_0 [label=Foo,
            shape=Mrecord];

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Record Field Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        label: str=None,
        *,
        name: str=None
        ) -> None:
        UniqueTreeNode.__init__(self, name=name)
        Attachable.__init__(self)
        if label is not None:
            label = str(label)
        self._label = label

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = '<{}>'.format(self._get_port_name())
        if self.label:
            result = '{} {}'.format(result, self.label)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def label(self) -> Optional[str]:
        return self._label



import collections
import enum
import math
import re
import uqbar.graphs  # noqa
from typing import Any, FrozenSet, Mapping, Optional, Union


class Attributes(collections.MutableMapping):
    """
    An attributes listing for a Graphviz graph, cluster, node or edge.

    ::

        >>> import uqbar.graphs
        >>> attributes = uqbar.graphs.Attributes(
        ...     'node',
        ...     color='black',
        ...     fillcolor='white',
        ...     shape='Mrecord',
        ...     style=['rounded', 'filled'],
        ...     )

    ::

        >>> print(format(attributes, 'graphviz'))
        [color=black,
            fillcolor=white,
            shape=Mrecord,
            style="rounded, filled"];

    """

    ### CLASS VARIABLES ###

    class Color(object):

        __slots__ = ('color',)

        def __init__(self, color) -> None:
            self.color = str(color)

        def __eq__(self, other) -> bool:
            return (
                type(self) == type(other) and
                self.color == other.color
                )

        def __repr__(self) -> str:
            return '<Color {!r}>'.format(self.color)

    class Mode(enum.Enum):
        CLUSTER = 1
        EDGE = 2
        GRAPH = 3
        NODE = 4
        TABLE = 5
        TABLE_CELL = 6

    class Point(object):

        __slots__ = ('x', 'y')

        def __init__(self, x, y) -> None:
            self.x = float(x)
            self.y = float(y)

        def __eq__(self, other) -> bool:
            return (
                type(self) == type(other) and
                self.x == other.x and
                self.y == other.y
                )

    __documentation_section__ = 'Core Classes'

    _arrow_types = frozenset(['box', 'circle', 'crow', 'diamond', 'dot',
        'ediamond', 'empty', 'halfopen', 'inv', 'invdot', 'invempty',
        'invodot', 'none', 'normal', 'obox', 'odiamond', 'odot', 'open', 'tee',
        'vee'])

    _cluster_modes = frozenset(['global', 'local', 'none'])

    _dir_types = frozenset(['back', 'both', 'forward', 'none'])

    _output_modes = frozenset(['breadthfirst', 'nodesfirst', 'edgesfirst'])

    _pack_modes = frozenset(['node', 'clust', 'graph'])

    _page_dirs = frozenset(['BL', 'BR', 'LB', 'LT', 'RB', 'RT', 'TL', 'TR'])

    _quad_types = frozenset(['fast', 'none', 'normal'])

    _rank_types = frozenset(['max', 'min', 'same', 'sink', 'source'])

    _rank_dirs = frozenset(['BT', 'LR', 'RL', 'TB'])

    _shapes = frozenset(['Mcircle', 'Mdiamond', 'Msquare', 'assembly', 'box',
        'box3d', 'cds', 'circle', 'component', 'cylinder', 'diamond',
        'doublecircle', 'doubleoctagon', 'egg', 'ellipse', 'fivepoverhang',
        'folder', 'hexagon', 'house', 'insulator', 'invhouse', 'invtrapezium',
        'invtriangle', 'larrow', 'lpromoter', 'none', 'note', 'noverhang',
        'octagon', 'oval', 'parallelogram', 'pentagon', 'plain', 'plaintext',
        'point', 'polygon', 'primersite', 'promoter', 'proteasesite',
        'proteinstab', 'rarrow', 'rect', 'rectangle', 'restrictionsite',
        'ribosite', 'rnastab', 'rpromoter', 'septagon', 'signature', 'square',
        'star', 'tab', 'terminator', 'threepoverhang', 'trapezium', 'triangle',
        'tripleoctagon', 'underline', 'utr', 'record', 'Mrecord'])

    _smooth_types = frozenset(['avg_dist', 'graph_dist', 'none', 'power_dist',
        'rng', 'spring', 'triangle'])

    _styles: FrozenSet[str] = frozenset()

    _word_pattern = re.compile('^\w+$')

    ### GRAPH OBJECT SPECIFICS ###

    _cluster_attributes = frozenset(['K', 'URL', 'area', 'bgcolor', 'color',
        'colorscheme', 'fillcolor', 'fontcolor', 'fontname', 'fontsize',
        'gradientangle', 'href', 'id', 'label', 'labeljust', 'labelloc',
        'layer', 'lheight', 'lp', 'lwidth', 'margin', 'nojustify', 'pencolor',
        'penwidth', 'peripheries', 'sortv', 'style', 'target', 'tooltip'])

    _cluster_styles = frozenset(['bold', 'dashed', 'dotted', 'filled',
        'rounded', 'solid', 'striped'])

    _edge_attributes = frozenset(['arrowhead', 'arrowsize', 'arrowtail',
        'color', 'colorscheme', 'comment', 'constraint', 'decorate', 'dir',
        'edgeURL', 'edgehref', 'edgetarget', 'edgetooltip', 'fillcolor',
        'fontcolor', 'fontname', 'fontsize', 'headURL', 'head_lp', 'headclip',
        'headhref', 'headlabel', 'headport', 'headtarget', 'headtooltip',
        'href', 'id', 'label', 'labelURL', 'labelangle', 'labeldistance',
        'labelfloat', 'labelfontcolor', 'labelfontname', 'labelfontsize',
        'labelhref', 'labeltarget', 'labeltooltip', 'layer', 'len', 'lhead',
        'lp', 'ltail', 'minlen', 'nojustify', 'penwidth', 'pos', 'samehead',
        'sametail', 'showboxes', 'style', 'tailURL', 'tail_lp', 'tailclip',
        'tailhref', 'taillabel', 'tailport', 'tailtarget', 'tailtooltip',
        'target', 'tooltip', 'weight'])

    _edge_styles = frozenset(['bold', 'dashed', 'dotted', 'solid'])

    _graph_attributes = frozenset(['Damping', 'K', 'URL', 'bb', 'bgcolor',
        'center', 'charset', 'clusterrank', 'color', 'colorscheme', 'comment',
        'compound', 'concentrate', 'defaultdist', 'dim', 'dimen',
        'diredgeconstraints', 'dpi', 'epsilon', 'esep', 'fontcolor',
        'fontname', 'fontnames', 'fontpath', 'fontsize', 'forcedlabels',
        'gradientangle', 'href', 'id', 'imagepath', 'inputscale', 'label',
        'label_scheme', 'labeljust', 'labelloc', 'landscape', 'layerlistsep',
        'layers', 'layerselect', 'layersep', 'layout', 'levels', 'levelsgap',
        'lheight', 'lp', 'lwidth', 'margin', 'maxiter', 'mclimit', 'mode',
        'model', 'mosek', 'newrank', 'nodesep', 'nojustify', 'normalize',
        'notranslate', 'nslimit', 'nslimit1', 'ordering', 'orientation',
        'outputorder', 'overlap', 'overlap_scaling', 'overlap_shrink', 'pack',
        'packmode', 'pad', 'page', 'pagedir', 'quadtree', 'quantum', 'rank',
        'rankdir', 'ranksep', 'remincross', 'repulsiveforce', 'resolution',
        'root', 'rotate', 'rotation', 'scale', 'searchsize', 'sep',
        'showboxes', 'size', 'smoothing', 'sortv', 'splines', 'start', 'style',
        'stylesheet', 'target', 'truecolor', 'viewport', 'voro_margin',
        'xdotversion', 'xlabel', 'xlp', 'penwidth'])

    _graph_styles = _cluster_styles

    _node_attributes = frozenset(['URL', 'area', 'color', 'colorscheme',
        'comment', 'distortion', 'fillcolor', 'fixedsize', 'fontcolor',
        'fontname', 'fontsize', 'gradientangle', 'group', 'height', 'href',
        'id', 'image', 'imagepos', 'imagescale', 'label', 'labelloc', 'layer',
        'margin', 'nojustify', 'ordering', 'orientation', 'penwidth',
        'peripheries', 'pin', 'pos', 'rects', 'regular', 'root',
        'samplepoints', 'shape', 'shapefile', 'showboxes', 'sides', 'skew',
        'sortv', 'style', 'target', 'tooltip', 'vertixes', 'width', 'z'])

    _node_styles = frozenset(['solid', 'dashed', 'dotted', 'bold', 'rounded',
        'diagonals', 'filled', 'striped', 'wedged'])

    ### HTML OBJECT SPECIFICS ###

    _table_attributes = frozenset(['align', 'bgcolor', 'border', 'cellborder',
        'cellpadding', 'cellspacing', 'color', 'columns', 'fixedsize',
        'gradientangle', 'height', 'href', 'id', 'rows', 'sides', 'style',
        'target', 'title', 'tooltip', 'valign', 'width'])

    _table_cell_attributes = frozenset(['align', 'balign', 'bgcolor', 'border',
        'cellpadding', 'cellspacing', 'color', 'colspan', 'fixedsize',
        'gradientangle', 'height', 'href', 'id', 'rowspan', 'sides', 'style',
        'target', 'title', 'tooltip', 'valign', 'width'])

    ### VALIDATORS ###

    _validators: Optional[Mapping[str, object]] = None

    ### INITIALIZER ###

    def __init__(
        self,
        mode: Union[str, 'uqbar.graphs.Attributes.Mode'],
        **kwargs,
        ) -> None:
        if not isinstance(mode, self.Mode):
            mode = self.Mode[str(mode).upper()]
        self._mode = mode
        self._attributes = self._validate_attributes(mode, **kwargs)

    ### SPECIAL METHODS ###

    def __delitem__(self, key: str) -> None:
        del(self._attributes[key])

    def __eq__(self, other) -> bool:
        return (
            type(self) == type(other) and
            self._mode == other._mode and
            self._attributes == other._attributes
            )

    def __format__(self, format_spec: str=None) -> str:
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        elif format_spec == 'html':
            return self.__format_html__()
        return str(self)

    def __format_graphviz__(self) -> str:
        if not self._attributes:
            return ''
        result = []
        attributes = sorted(self._attributes.items())
        for i, (key, value) in enumerate(attributes, 1):
            value = self._format_value(value).split('\n')
            value[0] = '{}={}'.format(key, value[0])
            if i < len(attributes):
                value[-1] += ','
            if i == 1:
                result.append(value.pop(0))
            result.extend('    ' + _ for _ in value)
        result[0] = '[' + result[0]
        result[-1] = result[-1] + '];'
        return '\n'.join(result)

    def __format_html__(self) -> str:
        if not self._attributes:
            return ''
        result = []
        for key, value in sorted(self._attributes.items()):
            value = self._format_value(value)
            if not value.startswith('"'):
                value = '"{}"'.format(value)
            result.append('{}={}'.format(key.upper(), value))
        return ' '.join(result)

    def __getitem__(self, key) -> Any:
        return self._attributes[key]

    def __iter__(self):
        return iter(self._attributes)

    def __len__(self):
        return len(self._attributes)

    def __setitem__(self, key, value):
        new_attributes = self._validate_attributes(
            self.mode, **{key: value})
        self._attributes.update(new_attributes)

    ### PRIVATE METHODS ###

    @classmethod
    def _format_value(cls, value) -> str:
        if isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, int):
            return str(value)
        elif isinstance(value, float):
            if not value % 1.0:
                value = math.floor(value)
            return str(value)
        elif isinstance(value, (list, tuple)):
            value = ', '.join(cls._format_value(x) for x in value)
            return cls._format_value(value)
        elif isinstance(value, cls.Color):
            return str(value.color)
        elif isinstance(value, str):
            if value.startswith('<') and value.endswith('>'):
                return value
            should_quote = False
            if not cls._word_pattern.match(value):
                should_quote = True
            elif value and value[0].isdigit():
                should_quote = True
            elif value.lower() in (
                'digraph',
                'edge',
                'graph',
                'node',
                'subgraph',
                ):
                should_quote = True
            if should_quote:
                value = value.replace('"', r'\"')
                value = '"{}"'.format(value)
            return value
        raise ValueError(value)

    @classmethod
    def _validate_arrow_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._arrow_types
        return value

    @classmethod
    def _validate_attributes(cls, mode, **kwargs):
        valid_attributes, valid_styles = {
            cls.Mode.CLUSTER: (cls._cluster_attributes, cls._cluster_styles),
            cls.Mode.EDGE: (cls._edge_attributes, cls._edge_styles),
            cls.Mode.GRAPH: (cls._graph_attributes, cls._graph_styles),
            cls.Mode.NODE: (cls._node_attributes, cls._node_styles),
            cls.Mode.TABLE: (cls._table_attributes, ()),
            cls.Mode.TABLE_CELL: (cls._table_cell_attributes, ()),
            }[mode]
        attributes = {}
        for key, value in kwargs.items():
            if key not in valid_attributes:
                raise ValueError(key)
            validators = cls._get_validators(mode)[key]
            if not isinstance(validators, tuple):
                validators = (validators,)
            for validator in validators:
                if isinstance(validator, str):
                    if str(value) == validator:
                        value = str(value)
                        break
                    continue
                elif isinstance(validator, type):
                    value = validator(value)
                    break
                else:
                    value = validator(value, valid_styles=valid_styles)
                    break
            attributes[key] = value
        return attributes

    @classmethod
    def _validate_cluster_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._cluster_modes
        return value

    @classmethod
    def _validate_color(cls, value, **kwargs):
        if isinstance(value, cls.Color):
            return value
        value = cls.Color(value)
        return value

    @classmethod
    def _validate_colors(cls, value, **kwargs):
        if isinstance(value, (cls.Color, int, str)):
            return cls._validate_color(value, **kwargs)
        assert len(value)
        value = tuple(cls._validate_color(_, **kwargs) for _ in value)
        if len(value) == 1:
            return value[0]
        return value

    @classmethod
    def _validate_dir_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._dir_types
        return value

    @classmethod
    def _validate_floats(cls, value, **kwargs):
        assert len(value)
        return tuple(float(_) for _ in value)

    @classmethod
    def _validate_output_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._output_modes
        return value

    @classmethod
    def _validate_pack_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._pack_modes
        return value

    @classmethod
    def _validate_page_dir(cls, value, **kwargs):
        value = str(value)
        assert value in cls._page_dirs
        return value

    @classmethod
    def _validate_point(cls, value, **kwargs):
        if isinstance(value, cls.Point):
            return value
        value = cls.Point(*value)
        return value

    @classmethod
    def _validate_points(cls, value_list, **kwargs):
        assert value_list
        return tuple(cls._validate_point(_, **kwargs) for _ in value_list)

    @classmethod
    def _validate_quad_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._quad_types
        return value

    @classmethod
    def _validate_rank_dir(cls, value, **kwargs):
        value = str(value)
        assert value in cls._rank_dirs
        return value

    @classmethod
    def _validate_rank_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._rank_types
        return value

    @classmethod
    def _validate_rect(cls, value, **kwargs):
        assert len(value) == 4
        value = tuple(float(_) for _ in value)
        return value

    @classmethod
    def _validate_shape(cls, value, **kwargs):
        value = str(value)
        assert value in cls._shapes
        return value

    @classmethod
    def _validate_smooth_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._smooth_types
        return value

    @classmethod
    def _validate_style(cls, value, valid_styles=None, **kwargs):
        value = str(value)
        assert value in valid_styles
        return value

    @classmethod
    def _validate_styles(cls, value, valid_styles=None, **kwargs):
        if isinstance(value, str):
            return cls._validate_style(
                value,
                valid_styles=valid_styles,
                **kwargs
                )
        assert value
        return tuple(cls._validate_style(
            _, valid_styles=valid_styles, **kwargs
            ) for _ in value)

    ### PUBLIC METHODS ###

    def copy(self):
        return type(self)(self.mode, **self._attributes.copy())

    ### PUBLIC PROPERTIES ###

    @property
    def mode(self):
        return self._mode

    @classmethod
    def _get_validators(cls, mode):
        if not cls._validators:
            cls._validators = {
                'Damping': float,
                'K': float,
                'URL': str,
                '_background': str,
                'align': ('center', 'left', 'right', 'text'),
                'area': float,
                'arrowhead': Attributes._validate_arrow_type,
                'arrowsize': float,
                'arrowtail': Attributes._validate_arrow_type,
                'balign': ('center', 'left', 'right', 'text'),
                'bb': Attributes._validate_rect,
                'bgcolor': Attributes._validate_colors,
                'border': float,
                'cellborder': float,
                'cellpadding': float,
                'cellspacing': float,
                'center': bool,
                'charset': str,
                'clusterrank': Attributes._validate_cluster_mode,
                'color': Attributes._validate_colors,
                'colorscheme': str,
                'colspan': int,
                'columns': int,
                'comment': str,
                'compound': bool,
                'concentrate': bool,
                'constraint': bool,
                'decorate': bool,
                'defaultdist': float,
                'dim': int,
                'dimen': int,
                'dir': Attributes._validate_dir_type,
                'diredgeconstraints': ('hier', bool),
                'distortion': float,
                'dpi': float,
                'edgeURL': str,
                'edgehref': str,
                'edgetarget': str,
                'edgetooltip': str,
                'epsilon': float,
                'esep': (float, Attributes._validate_point),
                'fillcolor': Attributes._validate_colors,
                'fixedsize': ('shape', bool),
                'fontcolor': Attributes._validate_color,
                'fontname': str,
                'fontnames': str,
                'fontpath': str,
                'fontsize': float,
                'forcelabels': bool,
                'gradientangle': int,
                'group': str,
                'headURL': str,
                'head_lp': Attributes._validate_point,
                'headclip': bool,
                'headhref': str,
                'headlabel': str,
                'headport': str,
                'headtarget': str,
                'headtooltip': str,
                'height': float,
                'href': str,
                'id': str,
                'image': str,
                'imagepath': str,
                'imagepos': str,
                'imagescale': ('width', 'height', 'both', bool),
                'inputscale': float,
                'label': str,
                'labelURL': str,
                'label_scheme': int,
                'labelangle': float,
                'labeldistance': float,
                'labelfloat': bool,
                'labelfontcolor': Attributes._validate_color,
                'labelfontname': str,
                'labelfontsize': float,
                'labelhref': str,
                'labeljust': str,
                'labelloc': str,
                'labeltarget': str,
                'labeltooltip': str,
                'landscape': bool,
                'layer': str,
                'layerlistsep': str,
                'layers': str,
                'layerselect': str,
                'layersep': str,
                'layout': str,
                'len': float,
                'levels': int,
                'levelsgap': float,
                'lhead': str,
                'lheight': float,
                'lp': Attributes._validate_point,
                'ltail': str,
                'lwidth': float,
                'margin': (float, Attributes._validate_point),
                'maxiter': int,
                'mclimit': float,
                'mindist': float,
                'minlen': int,
                'mode': str,
                'model': str,
                'mosek': bool,
                'newrank': bool,
                'nodesep': float,
                'nojustify': bool,
                'normalize': (float, bool),
                'notranslate': bool,
                'nslimit': float,
                'nslimit1': float,
                'ordering': str,
                'outputorder': Attributes._validate_output_mode,
                'overlap': ('scale', 'scalexy', 'compress', 'ipsep', 'prism', bool),
                'overlap_scaling': float,
                'overlap_shrink': bool,
                'pack': bool,
                'packmode': Attributes._validate_pack_mode,
                'pad': (float, Attributes._validate_point),
                'page': (float, Attributes._validate_point),
                'pagedir': Attributes._validate_page_dir,
                'pencolor': Attributes._validate_color,
                'penwidth': float,
                'peripheries': int,
                'pin': bool,
                'pos': Attributes._validate_point,
                'quadtree': (Attributes._validate_quad_type, bool),
                'quantum': float,
                'rank': Attributes._validate_rank_type,
                'rankdir': Attributes._validate_rank_dir,
                'ranksep': (float, Attributes._validate_floats),
                'ratio': ('fill', 'compress', 'expand', 'auto', float),
                'rects': Attributes._validate_rect,
                'regular': bool,
                'remincross': bool,
                'repulsiveforce': float,
                'resolution': float,
                'root': str,
                'rotate': int,
                'rotation': float,
                'rows': int,
                'rowspan': int,
                'samehead': str,
                'sametail': str,
                'samplepoints': int,
                'scale': (float, Attributes._validate_point),
                'searchsize': int,
                'sep': (float, Attributes._validate_point),
                'shape': Attributes._validate_shape,
                'shapefile': str,
                'showboxes': int,
                'sides': (int, str),
                'size': (float, Attributes._validate_point),
                'skew': float,
                'smoothing': Attributes._validate_smooth_type,
                'sortv': int,
                'splines': ('none', 'line', 'polyline', 'curved', 'ortho', 'spline', bool),
                'start': str,
                'style': Attributes._validate_styles,
                'stylesheet': str,
                'tailURL': str,
                'tail_lp': Attributes._validate_point,
                'tailclip': bool,
                'tailhref': str,
                'taillabel': str,
                'tailport': str,
                'tailtarget': str,
                'tailtooltip': str,
                'target': str,
                'title': str,
                'tooltip': str,
                'truecolor': bool,
                'valign': ('middle', 'bottom', 'top'),
                'vertices': Attributes._validate_points,
                'viewport': str,
                'voro_margin': float,
                'weight': int,
                'width': float,
                'xdotversion': str,
                'xlabel': str,
                'xlp': Attributes._validate_point,
                'z': float,
                }
        return cls._validators



from uqbar.containers.UniqueTreeNode import UniqueTreeNode


class VRule(UniqueTreeNode):
    """
    A Graphviz HTML vertical rule.

    ::

        >>> import uqbar.graphs
        >>> hrule = uqbar.graphs.VRule()
        >>> print(format(hrule, 'graphviz'))
        <VR/>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'HTML Classes'

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: str=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        return '<VR/>'



import abc
import argparse
import os
import pathlib
import typing
import uqbar.strings
from configparser import ConfigParser


class CLI(abc.ABC):
    '''
    Abstract base for CLI scripts.
    '''

    ### CLASS VARIABLES ###

    _colors = {
        'BLUE': '\033[94m',
        'RED': '\033[91m',
        'GREEN': '\033[92m',
        'END': '\033[0m',
        }

    config_name: str = '.uqbarrc'
    long_description: typing.Optional[str] = None
    short_description: typing.Optional[str] = None
    version: float = 1.0

    ### INITIALIZER ###

    def __init__(self):
        short_description = getattr(self, 'short_description', None)
        long_description = getattr(self, 'long_description', None)
        if long_description:
            long_description = uqbar.strings.normalize(long_description)
        parser = self._argument_parser = argparse.ArgumentParser(
            description=short_description,
            epilog=long_description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prog=self.program_name,
            )
        version = '%(prog)s {}'
        version = version.format(getattr(self, 'version', 1.0))
        parser.add_argument('--version', action='version', version=version)
        self._argument_parser = parser
        self._setup_argument_parser(parser)
        self._config_parser = None

    ### SPECIAL METHODS ###

    def __call__(self, arguments=None) -> None:
        r'''Calls developer script.

        Returns none.
        '''
        self._config_parser = self._read_config_files()
        if arguments is None:
            arguments = self.argument_parser.parse_args()
        else:
            if isinstance(arguments, str):
                arguments = arguments.split()
            elif not isinstance(arguments, (list, tuple)):
                raise ValueError
            arguments = self.argument_parser.parse_args(arguments)
        self._process_args(arguments)

    ### PRIVATE METHODS ###

    def _is_valid_path(self, path) -> bool:
        path = pathlib.Path(path)
        return path.exists() and path.is_dir()

    @abc.abstractmethod
    def _process_args(self, arguments):
        raise NotImplementedError

    def _read_config_files(self) -> ConfigParser:
        paths = []
        home_config = pathlib.Path().home() / self.config_name
        if home_config.exists():
            paths.append(home_config)
        path = pathlib.Path.cwd()
        while not path.joinpath(self.config_name).exists():
            path = path.parent
            if path.parent == path:
                break
        if path.joinpath(self.config_name).exists():
            if path.joinpath(self.config_name) not in paths:
                paths.append(path.joinpath(self.config_name))
        config_parser = ConfigParser()
        config_parser.read([str(_) for _ in paths])
        return config_parser

    @abc.abstractmethod
    def _setup_argument_parser(self, parser: argparse.ArgumentParser):
        raise NotImplementedError

    def _validate_path(self, path):
        message = '{!r} is not a valid directory.'
        message = message.format(path)
        error = argparse.ArgumentTypeError(message)
        path = os.path.abspath(path)
        if not self._is_valid_path(path):
            raise error
        return os.path.relpath(path)

    ### PUBLIC PROPERTIES ###

    @property
    def argument_parser(self):
        r'''The script's instance of argparse.ArgumentParser.
        '''
        return self._argument_parser

    @property
    def formatted_help(self):
        r'''Formatted help of developer script.
        '''
        return self._argument_parser.format_help()

    @property
    def formatted_usage(self):
        r'''Formatted usage of developer script.
        '''
        return self._argument_parser.format_usage()

    @property
    def formatted_version(self):
        r'''Formatted version of developer script.
        '''
        return self._argument_parser.format_version()

    @property
    def program_name(self):
        r'''The name of the script, callable from the command line.
        '''
        name = '-'.join(
            word.lower() for word in
            uqbar.strings.delimit_words(type(self).__name__)
            )
        return name



"""
Uqbar Sphinx API generation extension.

Install by adding ``'uqbar.sphinx.api'`` to the ``extensions`` list in your
Sphinx configuration.

This extension provides the following configuration values which correspond to
the initialization arguments to the :py:class:`uqbar.apis.APIBuilder` class.

- ``uqbar_api_directory_name`` (default: ``api``)
- ``uqbar_api_document_empty_modules`` (default: ``False``)
- ``uqbar_api_document_private_members`` (default: ``False``)
- ``uqbar_api_document_private_modules`` (default: ``False``)
- ``uqbar_api_member_documenter_classes`` (default: ``None``)
- ``uqbar_api_module_documenter_class`` (default: ``None``)
- ``uqbar_api_root_documenter_class`` (default: ``None``)
- ``uqbar_api_source_paths`` (default: ``[]``)
- ``uqbar_api_title`` (default: ``API``)

reStructuredText source files will be generated for the modules given by
``uqbar_api_source_paths`` in the directory ``uqbar_api_directory_name``
relative to your Sphinx source directory.
"""
import importlib
import pathlib
import sphinx.application  # type: ignore
import types
import uqbar.apis
from typing import List  # noqa


def on_builder_inited(app: sphinx.application.Sphinx):
    """
    Hooks into Sphinx's ``builder-inited`` event.

    Builds out the ReST API source.
    """
    config = app.builder.config

    target_directory = (
        pathlib.Path(app.builder.env.srcdir) /
        config.uqbar_api_directory_name
        )

    initial_source_paths: List[str] = []
    source_paths = config.uqbar_api_source_paths
    for source_path in source_paths:
        if isinstance(source_path, types.ModuleType):
            if hasattr(source_path, '__path__'):
                initial_source_paths.extend(getattr(source_path, '__path__'))
            else:
                initial_source_paths.extend(source_path.__file__)
            continue
        try:
            module = importlib.import_module(source_path)
            if hasattr(module, '__path__'):
                initial_source_paths.extend(getattr(module, '__path__'))
            else:
                initial_source_paths.append(module.__file__)
        except ImportError:
            initial_source_paths.append(source_path)

    root_documenter_class = config.uqbar_api_root_documenter_class
    if isinstance(root_documenter_class, str):
        module_name, _, class_name = root_documenter_class.rpartition('.')
        module = importlib.import_module(module_name)
        root_documenter_class = getattr(module, class_name)

    module_documenter_class = config.uqbar_api_module_documenter_class
    if isinstance(module_documenter_class, str):
        module_name, _, class_name = module_documenter_class.rpartition('.')
        module = importlib.import_module(module_name)
        module_documenter_class = getattr(module, class_name)

    member_documenter_classes = config.uqbar_api_member_documenter_classes or []
    for i, member_documenter_class in enumerate(member_documenter_classes):
        if isinstance(member_documenter_class, str):
            module_name, _, class_name = member_documenter_class.rpartition('.')
            module = importlib.import_module(module_name)
            member_documenter_classes[i] = getattr(module, class_name)
    if not member_documenter_classes:
        member_documenter_classes = None

    api_builder = uqbar.apis.APIBuilder(
        initial_source_paths=initial_source_paths,
        target_directory=target_directory,
        document_empty_modules=config.uqbar_api_document_empty_modules,
        document_private_members=config.uqbar_api_document_private_members,
        document_private_modules=config.uqbar_api_document_private_modules,
        member_documenter_classes=member_documenter_classes,
        module_documenter_class=module_documenter_class,
        root_documenter_class=root_documenter_class,
        title=config.uqbar_api_title,
        )
    api_builder()


def setup(app: sphinx.application.Sphinx):
    """
    Sets up Sphinx extension.
    """
    app.connect('builder-inited', on_builder_inited)
    app.add_config_value('uqbar_api_directory_name', 'api', 'env')
    app.add_config_value('uqbar_api_document_empty_modules', False, 'env')
    app.add_config_value('uqbar_api_document_private_members', False, 'env')
    app.add_config_value('uqbar_api_document_private_modules', False, 'env')
    app.add_config_value('uqbar_api_member_documenter_classes', None, 'env')
    app.add_config_value('uqbar_api_module_documenter_class', None, 'env')
    app.add_config_value('uqbar_api_root_documenter_class', None, 'env')
    app.add_config_value('uqbar_api_source_paths', [], 'env')
    app.add_config_value('uqbar_api_title', 'API', 'env')



import os
import pathlib


class DirectoryChange:
    """
    A context manager for temporarily changing the current working directory.

    This context manager is re-entrant.
    """

    ### INITIALIZER ###

    def __init__(self, directory, verbose=False):
        directory = pathlib.Path(directory).resolve().absolute()
        if not directory.is_dir():
            directory = directory.parent
        self._directory = directory
        self._directory_stack = []
        self._verbose = bool(verbose)

    ### SPECIAL METHODS ###

    def __enter__(self):
        if self.verbose:
            print('Changing directory to {} ...'.format(self.directory))
        self._directory_stack.append(pathlib.Path.cwd())
        os.chdir(str(self._directory))
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        original_directory = self._directory_stack.pop()
        if self.verbose:
            print('Returning to {} ...'.format(original_directory))
        os.chdir(str(original_directory))

    ### PUBLIC PROPERTIES ###

    @property
    def directory(self):
        return self._directory

    @property
    def verbose(self):
        return self._verbose



"""
Tools for IO and file-system manipulation.
"""

import collections
import os
import pathlib
from typing import Generator, List, Optional, Sequence, Tuple, Union

from .DirectoryChange import DirectoryChange  # noqa
from .Profiler import Profiler  # noqa
from .RedirectedStreams import RedirectedStreams  # noqa
from .Timer import Timer  # noqa


def find_common_prefix(
    paths: Sequence[Union[str, pathlib.Path]]
    ) -> Optional[pathlib.Path]:
    """
    Find the common prefix of two or more paths.

    ::

        >>> one = pathlib.Path('foo/bar/baz')
        >>> two = pathlib.Path('foo/quux/biz')
        >>> three = pathlib.Path('foo/quux/wuux')

    ::

        >>> import uqbar.io
        >>> str(uqbar.io.find_common_prefix([one, two, three]))
        'foo'

    :param paths: paths to inspect
    """
    counter: collections.Counter = collections.Counter()
    for path in paths:
        path = pathlib.Path(path)
        counter.update([path])
        counter.update(path.parents)
    valid_paths = sorted([
        path for path, count in counter.items()
        if count >= len(paths)
        ], key=lambda x: len(x.parts))
    if valid_paths:
        return valid_paths[-1]
    return None


def find_executable(name: str, flags=os.X_OK) -> List[str]:
    r'''Finds executable `name`.

    Similar to Unix ``which`` command.

    Returns list of zero or more full paths to `name`.
    '''
    result = []
    extensions = [
        x
        for x in os.environ.get('PATHEXT', '').split(os.pathsep)
        if x
        ]
    path = os.environ.get('PATH', None)
    if path is None:
        return []
    for path in os.environ.get('PATH', '').split(os.pathsep):
        path = os.path.join(path, name)
        if os.access(path, flags):
            result.append(path)
        for extension in extensions:
            path_extension = path + extension
            if os.access(path_extension, flags):
                result.append(path_extension)
    return result


def relative_to(
    source_path: Union[str, pathlib.Path],
    target_path: Union[str, pathlib.Path],
    ) -> pathlib.Path:
    """
    Generates relative path from ``source_path`` to ``target_path``.

    Handles the case of paths without a common prefix.

    ::

        >>> source = pathlib.Path('foo/bar/baz')
        >>> target = pathlib.Path('foo/quux/biz')

    ::

        >>> target.relative_to(source)
        Traceback (most recent call last):
        ...
        ValueError: 'foo/quux/biz' does not start with 'foo/bar/baz'

    ::

        >>> import uqbar.io
        >>> str(uqbar.io.relative_to(source, target))
        '../../quux/biz'

    :param source_path: the source path
    :param target_path: the target path
    """
    source_path = pathlib.Path(source_path).absolute()
    if source_path.is_file():
        source_path = source_path.parent
    target_path = pathlib.Path(target_path).absolute()
    common_prefix = find_common_prefix([source_path, target_path])
    if not common_prefix:
        raise ValueError('No common prefix')
    source_path = source_path.relative_to(common_prefix)
    target_path = target_path.relative_to(common_prefix)
    result = pathlib.Path(*['..'] * len(source_path.parts))
    return result / target_path


def walk(
    root_path: Union[str, pathlib.Path],
    top_down: bool=True,
    ) -> Generator[
        Tuple[pathlib.Path, Sequence[pathlib.Path], Sequence[pathlib.Path]],
        None, None]:
    """
    Walks a directory tree.

    Like :py:func:`os.walk` but yielding instances of :py:class:`pathlib.Path`
    instead of strings.

    :param root_path: foo
    :param top_down: bar
    """
    root_path = pathlib.Path(root_path)
    directory_paths, file_paths = [], []
    for path in sorted(root_path.iterdir()):
        if path.is_dir():
            directory_paths.append(path)
        else:
            file_paths.append(path)
    if top_down:
        yield root_path, directory_paths, file_paths
    for directory_path in directory_paths:
        yield from walk(directory_path, top_down=top_down)
    if not top_down:
        yield root_path, directory_paths, file_paths


def write(
    contents: str,
    path: Union[str, pathlib.Path],
    verbose: bool=False,
    ) -> bool:
    """
    Writes ``contents`` to ``path``.

    Checks if ``path`` already exists and only write out new contents if the
    old contents do not match.

    Creates any intermediate missing directories.

    :param contents: the file contents to write
    :param path: the path to write to
    :param verbose: whether to print output
    """
    path = pathlib.Path(path)
    if path.exists():
        with path.open('r') as file_pointer:
            old_contents = file_pointer.read()
        if old_contents == contents:
            if verbose:
                print('preserved: {}'.format(path))
            return False
        else:
            with path.open('w') as file_pointer:
                file_pointer.write(contents)
            if verbose:
                print('rewrote: {}'.format(path))
            return True
    elif not path.exists():
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        with path.open('w') as file_pointer:
            file_pointer.write(contents)
        if verbose:
            print('wrote: {}'.format(path))
    return True



"""
Tools for string manipulation.
"""

import textwrap
import unidecode  # type: ignore
from typing import Generator


def delimit_words(string: str) -> Generator[str, None, None]:
    """
    Delimit a string at word boundaries.
    """
    wordlike_characters = ('<', '>', '!')
    current_word = ''
    for character in string:
        if (
            not character.isalpha() and
            not character.isdigit() and
            character not in wordlike_characters
            ):
            if current_word:
                yield current_word
                current_word = ''
        elif not current_word:
            current_word += character
        elif character.isupper():
            if current_word[-1].isupper():
                current_word += character
            else:
                yield current_word
                current_word = character
        elif character.islower():
            if current_word[-1].isalpha():
                current_word += character
            else:
                yield current_word
                current_word = character
        elif character.isdigit():
            if current_word[-1].isdigit():
                current_word += character
            else:
                yield current_word
                current_word = character
        elif character in wordlike_characters:
            if current_word[-1] in wordlike_characters:
                current_word += character
            else:
                yield current_word
                current_word = character
    if current_word:
        yield current_word


def normalize(string: str) -> str:
    """
    Normalizes whitespace.

    Strips leading and trailing blank lines, dedents, and removes trailing
    whitespace from the result.
    """
    string = string.replace('\t', '    ')
    lines = string.split('\n')
    while lines and (not lines[0] or lines[0].isspace()):
        lines.pop(0)
    while lines and (not lines[-1] or lines[-1].isspace()):
        lines.pop()
    for i, line in enumerate(lines):
        lines[i] = line.rstrip()
    string = '\n'.join(lines)
    string = textwrap.dedent(string)
    return string


def to_dash_case(string: str) -> str:
    """
    Convert a string to dash-delimited words.

    ::

        >>> import uqbar.strings
        >>> string = 'Tô Đặc Biệt Xe Lửa'
        >>> print(uqbar.strings.to_dash_case(string))
        to-dac-biet-xe-lua

    ::

        >>> string = 'alpha.beta.gamma'
        >>> print(uqbar.strings.to_dash_case(string))
        alpha-beta-gamma

    """
    string = unidecode.unidecode(string)
    words = (_.lower() for _ in delimit_words(string))
    string = '-'.join(words)
    return string


def to_snake_case(string: str) -> str:
    """
    Convert a string to underscore-delimited words.

    ::

        >>> import uqbar.strings
        >>> string = 'Tô Đặc Biệt Xe Lửa'
        >>> print(uqbar.strings.to_snake_case(string))
        to_dac_biet_xe_lua

    ::

        >>> string = 'alpha.beta.gamma'
        >>> print(uqbar.strings.to_snake_case(string))
        alpha_beta_gamma

    """
    string = unidecode.unidecode(string)
    words = (_.lower() for _ in delimit_words(string))
    string = '_'.join(words)
    return string



"""
Uqbar Sphinx inheritance graph extension.

Install by adding ``'uqbar.sphinx.inheritance'`` to the ``extensions`` list in
your Sphinx configuration.

Invoke by adding the following directive to your ReST source:

::

    .. inheritance-graph:: some.python.path

This will generate Graphviz output mediated by a
:py:class:`uqbar.apis.InheritanceGraph` instance. You can specify lineage paths
by adding a ``:lineage:`` option to the directive:

::

    .. inheritance-graph:: some.python.path
       :lineage: some.lineage.path

Both the main argument to the directive and the lineage option can take
multiple paths, separated by spaces:

::

    .. inheritance-graph:: path.one path.two
       :lineage: path.three path.four path.five

"""
import math
import os
import pathlib
import subprocess
import uqbar.apis
from docutils.nodes import General, Element, Node, NodeVisitor, SkipNode  # type: ignore
from docutils.parsers.rst import Directive, directives  # type: ignore
from sphinx.ext.graphviz import render_dot_html, render_dot_latex  # type: ignore
from typing import List, Mapping


class inheritance_diagram(General, Element):
    """
    A docutils node to use as a placeholder for the inheritance diagram.
    """

    __documentation_ignore_inherited__ = True


class InheritanceDiagram(Directive):
    """
    Runs when the inheritance_diagram directive is first encountered.
    """
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'lineage': directives.unchanged}

    __documentation_ignore_inherited__ = True

    def run(self) -> List[Node]:
        node = inheritance_diagram()
        node.document = self.state.document
        package_paths = self.arguments[0].split()
        lineage_paths = self.options.get('lineage', '').split() or None
        # Create the graph
        try:
            graph = uqbar.apis.InheritanceGraph(
                package_paths=package_paths,
                lineage_paths=lineage_paths,
                )
        except Exception as error:
            warning = node.document.reporter.warning(
                error.args[0],
                line=self.lineno,
                )
            return [warning]
        if not graph.classes:
            return []
        # Create xref nodes for each target of the graph's image map and add
        # them to the doc tree so that Sphinx can resolve the references to
        # real URLs later.  These nodes will eventually be removed from the
        # doctree after we're done with them.
        env = self.state.document.settings.env
        class_role = env.get_domain('py').role('class')
        for class_ in graph.classes:
            url_name = class_.__name__
            if class_.__module__ not in ('__builtins__', 'builtins'):
                url_name = class_.__module__ + '.' + url_name
            refnodes, _ = class_role(
                'class', ':class:`{}`'.format(url_name),
                url_name, 0, self.state,
                )
            node.extend(refnodes)
        # Store the graph object so we can use it to generate the dot file
        # later on
        node['graph'] = graph
        return [node]


def build_urls(
    self: NodeVisitor,
    node: inheritance_diagram,
    ) -> Mapping[str, str]:
    """
    Builds a mapping of class paths to URLs.
    """
    current_filename = self.builder.current_docname + self.builder.out_suffix
    urls = {}
    for child in node:
        # Another document
        if child.get('refuri') is not None:
            uri = child.get('refuri')
            package_path = child['reftitle']
            if uri.startswith('http'):
                _, _, package_path = uri.partition('#')
            else:
                uri = (
                    pathlib.Path('..') /
                    pathlib.Path(current_filename).parent /
                    pathlib.Path(uri)
                    )
                uri = str(uri).replace(os.path.sep, '/')
            urls[package_path] = uri
        # Same document
        elif child.get('refid') is not None:
            urls[child['reftitle']] = '../' + current_filename + '#' + child.get('refid')
    return urls


def html_visit_inheritance_diagram(
    self: NodeVisitor,
    node: inheritance_diagram,
    ) -> None:
    """
    Builds HTML output from an :py:class:`~uqbar.sphinx.inheritance.inheritance_diagram` node.
    """
    inheritance_graph = node['graph']
    urls = build_urls(self, node)
    graphviz_graph = inheritance_graph.build_graph(urls)
    dot_code = format(graphviz_graph, 'graphviz')
    # TODO: We can perform unflattening here
    aspect_ratio = inheritance_graph.aspect_ratio
    if aspect_ratio:
        aspect_ratio = math.ceil(math.sqrt(aspect_ratio[1] / aspect_ratio[0]))
    if aspect_ratio > 1:
        process = subprocess.Popen(
            [
                'unflatten',
                '-l', str(aspect_ratio),
                '-c', str(aspect_ratio),
                '-f',
                ],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
        stdout, stderr = process.communicate(dot_code.encode())
        dot_code = stdout.decode()
    render_dot_html(self, node, dot_code, {}, 'inheritance', 'inheritance')
    raise SkipNode


def latex_visit_inheritance_diagram(
    self: NodeVisitor,
    node: inheritance_diagram,
    ) -> None:
    """
    Builds LaTeX output from an :py:class:`~uqbar.sphinx.inheritance.inheritance_diagram` node.
    """
    inheritance_graph = node['graph']
    graphviz_graph = inheritance_graph.build_graph()
    graphviz_graph.attributes['size'] = 6.0
    dot_code = format(graphviz_graph, 'graphviz')
    render_dot_latex(self, node, dot_code, {}, 'inheritance')
    raise SkipNode


def skip(self: NodeVisitor, node: inheritance_diagram) -> None:
    """
    Skip generating output, for non-supported builders.
    """
    raise SkipNode


def setup(app) -> None:
    """
    Sets up Sphinx extension.
    """
    app.setup_extension('sphinx.ext.graphviz')
    app.add_node(
        inheritance_diagram,
        html=(html_visit_inheritance_diagram, None),
        latex=(latex_visit_inheritance_diagram, None),
        man=(skip, None),
        texinfo=(skip, None),
        text=(skip, None),
        )
    app.add_directive('inheritance-diagram', InheritanceDiagram)



"""
Specialized container classes.
"""

from .DependencyGraph import DependencyGraph  # noqa
from .UniqueTreeNode import UniqueTreeNode  # noqa
from .UniqueTreeContainer import UniqueTreeContainer  # noqa



import pathlib


class RootDocumenter:
    """
    A basic root documenter.

    This documenter generates only one reStructuredText document: the root API
    index page, which contains information about all of the modules traversed
    by an :py:class:`~uqbar.apis.APIBuilder`.

    Output is a basic `toctree` directive.

    ::

        >>> import uqbar.apis
        >>> documenter = uqbar.apis.RootDocumenter(
        ...     module_documenters=[
        ...         uqbar.apis.ModuleDocumenter('uqbar.io'),
        ...         uqbar.apis.ModuleDocumenter('uqbar.strings'),
        ...         ],
        ...     )
        >>> print(str(documenter))
        API
        ===
        <BLANKLINE>
        .. toctree::
        <BLANKLINE>
           uqbar/io/index
           uqbar/strings
        <BLANKLINE>

    .. tip::

       Subclass :py:class:`~uqbar.apis.RootDocumenter` to
       implement your own custom module documentation output.
       You'll need to provide your desired reStructuredText output
       via an overridden
       :py:meth:`~uqbar.apis.RootDocumenter.RootDocumenter.__str__`
       implementation.

       See :py:class:`~uqbar.apis.SummarizingRootDocumenter` for an example.

    :param module_documenters: a list of of documenters for modules and
        packages of the root documenter; these are generated by an
        :py:class:`~uqbar.apis.APIBuilder` instance rather than the module
        documenter directly
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Documenters'

    ### INITIALIZER ###

    def __init__(self, module_documenters=None, title='API'):
        import uqbar.apis
        if module_documenters is not None:
            assert all(isinstance(_, uqbar.apis.ModuleDocumenter)
                for _ in module_documenters), module_documenters
            module_documenters = tuple(module_documenters)
        self._module_documenters = module_documenters or ()
        self._title = title

    ### SPECIAL METHODS ###

    def __str__(self):
        result = [self.title, '=' * len(self.title), '']
        if self.module_documenters:
            result.extend(['.. toctree::', ''])
            for module_documenter in self.module_documenters:
                path = module_documenter.package_path.replace('.', '/')
                if module_documenter.is_package:
                    path = '{}/index'.format(path)
                result.append('   {}'.format(path))
            result.append('')
        return '\n'.join(result)

    ### PUBLIC PROPERTIES ###

    @property
    def documentation_path(self):
        return pathlib.Path('index.rst')

    @property
    def module_documenters(self):
        return self._module_documenters

    @property
    def title(self):
        return self._title



import enum
import inspect
from typing import List
from uqbar.apis.ClassDocumenter import ClassDocumenter


class SummarizingClassDocumenter(ClassDocumenter):
    """
    A summarizing class documenter.

    Organizes class members by category, separated by category title and
    horizontal rule.

    Categories include:

    -  Special methods
    -  Methods
    -  Class and static methods
    -  Read/write properties
    -  Read-only properties

    ::

        >>> import uqbar.apis
        >>> path = 'uqbar.apis.SummarizingClassDocumenter.SummarizingClassDocumenter'
        >>> documenter = uqbar.apis.SummarizingClassDocumenter(path)
        >>> documentation = str(documenter)
        >>> print(documentation)
        .. autoclass:: SummarizingClassDocumenter
        <BLANKLINE>
           .. autosummary::
              :nosignatures:
        <BLANKLINE>
              ignored_special_methods
        <BLANKLINE>
           .. autosummary::
              :nosignatures:
        <BLANKLINE>
              __str__
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Special methods
              :class: class-header
        <BLANKLINE>
           .. automethod:: SummarizingClassDocumenter.__str__
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Class & static methods
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. automethod:: SummarizingClassDocumenter.validate_client
        <BLANKLINE>
           .. raw:: html
        <BLANKLINE>
              <hr/>
        <BLANKLINE>
           .. rubric:: Read-only properties
              :class: class-header
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.client
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.documentation_section
        <BLANKLINE>
           .. container:: inherited
        <BLANKLINE>
              .. autoattribute:: SummarizingClassDocumenter.package_path

    :param package_path: the module path and name of the member to document
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Documenters'

    ignored_special_methods = (
        '__dict__',
        '__getattribute__',
        '__getnewargs__',
        '__getstate__',
        '__init__',
        '__reduce__',
        '__reduce_ex__',
        '__setstate__',
        '__sizeof__',
        '__subclasshook__',
        'fromkeys',
        'pipe_cloexec',
        )

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        name = getattr(self.client, '__name__')
        if issubclass(self.client, Exception):  # type: ignore
            return '.. autoexception:: {}'.format(name)
        attributes = self._classify_class_attributes()
        (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
            ) = attributes
        result = [
            '.. autoclass:: {}'.format(name),
            ]
        if issubclass(self.client, enum.Enum):  # type: ignore
            result.extend([
                '   :members:',
                '   :undoc-members:',
            ])
        result.extend(self._build_member_autosummary(attributes))
        result.extend(self._build_attribute_section(
            special_methods,
            'automethod',
            'Special methods',
            ))
        result.extend(self._build_attribute_section(
            methods,
            'automethod',
            'Methods',
            ))
        result.extend(self._build_attribute_section(
            sorted(class_methods + static_methods, key=lambda x: x.name),
            'automethod',
            'Class & static methods',
            ))
        result.extend(self._build_attribute_section(
            readwrite_properties,
            'autoattribute',
            'Read/write properties',
            ))
        result.extend(self._build_attribute_section(
            readonly_properties,
            'autoattribute',
            'Read-only properties',
            ))
        return '\n'.join(result)

    ### PRIVATE METHODS ###

    def _build_attribute_section(
        self,
        attributes,
        directive: str,
        title: str,
    ) -> List[str]:
        result: List[str] = []
        if not attributes:
            return result
        result.extend([
            '',
            '   .. raw:: html',
            '',
            '      <hr/>',
            '',
            '   .. rubric:: {}'.format(title),
            '      :class: class-header',
            ])
        for attribute in attributes:
            result.append('')
            autodoc_directive = '   .. {}:: {}.{}'.format(
                directive, getattr(self.client, '__name__'), attribute.name)
            if attribute.defining_class is self.client:
                result.append(autodoc_directive)
            else:
                result.append('   .. container:: inherited')
                result.append('')
                result.append('   {}'.format(autodoc_directive))
        return result

    def _build_member_autosummary(self, attributes) -> List[str]:
        result: List[str] = []
        if not attributes:
            return result
        for attribute_section in attributes:
            attribute_section = sorted(
                attribute for attribute in attribute_section
                if attribute.defining_class is self.client
            )
            if not attribute_section:
                continue
            result.extend([
                '',
                '   .. autosummary::',
                '      :nosignatures:',
                '',
            ])
            for attribute in attribute_section:
                result.append('      {}'.format(attribute.name))
        return result

    def _classify_class_attributes(self):
        class_methods = []
        data = []
        methods = []
        readonly_properties = []
        readwrite_properties = []
        special_methods = []
        static_methods = []
        attrs = inspect.classify_class_attrs(self.client)
        for attr in attrs:
            if attr.defining_class is object:
                continue
            elif (
                getattr(self.client, '__documentation_ignore_inherited__', None) and
                attr.defining_class is not self.client
            ):
                continue
            if attr.kind == 'method':
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith('__'):
                        special_methods.append(attr)
                    elif not attr.name.startswith('_'):
                        methods.append(attr)
            elif attr.kind == 'class method':
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith('__'):
                        special_methods.append(attr)
                    elif not attr.name.startswith('_'):
                        class_methods.append(attr)
            elif attr.kind == 'static method':
                if attr.name not in self.ignored_special_methods:
                    if attr.name.startswith('__'):
                        special_methods.append(attr)
                    elif not attr.name.startswith('_'):
                        static_methods.append(attr)
            elif attr.kind == 'property' and not attr.name.startswith('_'):
                if attr.object.fset is None:
                    readonly_properties.append(attr)
                else:
                    readwrite_properties.append(attr)
            elif attr.kind == 'data' and not attr.name.startswith('_') \
                and attr.name not in getattr(self.client, '__slots__', ()):
                data.append(attr)
        class_methods = tuple(sorted(class_methods))
        data = tuple(sorted(data))
        methods = tuple(sorted(methods))
        readonly_properties = tuple(sorted(readonly_properties))
        readwrite_properties = tuple(sorted(readwrite_properties))
        special_methods = tuple(sorted(special_methods))
        static_methods = tuple(sorted(static_methods))
        result = (
            class_methods,
            data,
            methods,
            readonly_properties,
            readwrite_properties,
            special_methods,
            static_methods,
            )
        return result



#!/usr/bin/env python
import pathlib
import setuptools


package_name = 'uqbar'


def read_version():
    root_path = pathlib.Path(__file__).parent
    version_path = root_path / package_name / '_version.py'
    with version_path.open() as file_pointer:
        file_contents = file_pointer.read()
    local_dict = {}
    exec(file_contents, None, local_dict)
    return local_dict['__version__']


description = 'Tools for building documentation with Sphinx, Graphviz and LaTeX'

classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: MacOS',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3.6',
    ]

install_requires = [
    'Sphinx>=1.6.0',
    'Unidecode>=0.4.21',
    'sphinx-autodoc-typehints>=1.3.0',
    'sphinx-rtd-theme>=0.2.4',
    ]

keywords = [
    'sphinx',
    'graphviz',
    'latex',
    'documentation',
    ]

with open('README.md', 'r') as file_pointer:
    long_description = file_pointer.read()

version = read_version()

if __name__ == '__main__':
    setuptools.setup(
        author='Josiah Wolf Oberholtzer',
        author_email='josiah.oberholtzer@gmail.com',
        classifiers=classifiers,
        description=description,
        include_package_data=True,
        install_requires=install_requires,
        keywords=keywords,
        long_description=description,
        name=package_name,
        packages=[package_name],
        url='https://github.com/josiah-wolf-oberholtzer/uqbar',
        version=version,
        zip_safe=False,
        )



import abc
import importlib


class MemberDocumenter(metaclass=abc.ABCMeta):
    """
    Abstract base class for module member documenters.

    Member documenters generate reStructuredText output for classes, functions
    and other objects defined in modules. They also implement the logic for
    identifying which of those objects should be selected for documenting via
    their
    :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.validate_client`
    method. Module documenters loop over their client module's dict and try
    each of their member documenter class' validation method against each dict
    member in turn in order to identify documentable content and associate it
    with a member documenter class.

    .. tip::

       Subclass :py:class:`~uqbar.apis.MemberDocumenter` to implement your own
       custom module member documentation output. You'll need to override
       :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.validate_client`
       with custom identification logic, and provide your desired
       reStructuredText output via an overridden
       :py:meth:`~uqbar.apis.MemberDocumenter.MemberDocumenter.__str__`
       implementation.

    :param package_path: the module path and name of the member to document
    """

    ### INITIALIZER ###

    def __init__(self, package_path: str) -> None:
        module_path, _, client_name = package_path.rpartition('.')
        module = importlib.import_module(module_path)
        client = getattr(module, client_name)
        if not self.validate_client(client, module_path):
            message = 'Unexpected object: {}'.format(type(client))
            raise ValueError(message)
        self._client = client
        self._package_path = package_path

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @classmethod
    @abc.abstractmethod
    def validate_client(cls, client: object, module_path: str) -> bool:
        """
        Check if the member documenter can document the provided client.

        :param client: the potential client to validate
        :param module_path: the package path of the module where the client was encountered
        """
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def client(self) -> object:
        return self._client

    @property
    @abc.abstractmethod
    def documentation_section(self) -> str:
        raise NotImplementedError

    @property
    def package_path(self) -> str:
        return self._package_path



from uqbar.containers import UniqueTreeContainer
from typing import Tuple  # noqa


class RecordGroup(UniqueTreeContainer):
    """ A Graphviz record field group.

    ::

        >>> import uqbar.graphs
        >>> group = uqbar.graphs.RecordGroup()
        >>> group.extend([
        ...     uqbar.graphs.RecordField(),
        ...     uqbar.graphs.RecordGroup([
        ...         uqbar.graphs.RecordField(),
        ...         uqbar.graphs.RecordField(),
        ...         ]),
        ...     uqbar.graphs.RecordField(),
        ...     ])
        >>> print(format(group, 'graphviz'))
        { <f_0> | { <f_0> | <f_0> } | <f_0> }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Record Field Classes'

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        name: str=None
        ) -> None:
        UniqueTreeContainer.__init__(self, name=name, children=children)

    ### SPECIAL METHODS ###

    def __format__(self, format_spec=None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == 'graphviz':
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = ' | '.join(
            _ for _ in (format(_, 'graphviz') for _ in self) if _
            )
        if result:
            result = '{{ {} }}'.format(result)
        return result

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs
        return (
            uqbar.graphs.RecordField,
            uqbar.graphs.RecordGroup,
            )



import sys


class RedirectedStreams:
    """
    A context manager for capturing ``stdout`` and ``stderr`` output.

    ..  container:: example

        ::

            >>> import uqbar.io
            >>> import io

        ::

            >>> string_io = io.StringIO()
            >>> with uqbar.io.RedirectedStreams(stdout=string_io):
            ...     print("hello, world!")
            ...

        ::

            >>> result = string_io.getvalue()
            >>> string_io.close()
            >>> print(result)
            hello, world!

    This context manager is not reentrant. Use a separate instance when nesting
    multiple timers.
    """

    ### INITIALIZER ###

    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    ### SPECIAL METHODS ###

    def __enter__(self):
        self._old_stdout, self._old_stderr = sys.stdout, sys.stderr
        self._old_stdout.flush()
        self._old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self._stdout.flush()
            self._stderr.flush()
        finally:
            sys.stdout = self._old_stdout
            sys.stderr = self._old_stderr

    ### PUBLIC PROPERTIES ###

    @property
    def stderr(self):
        return self._stderr

    @property
    def stdout(self):
        return self._stdout
